import numpy as np
import warnings
from . import Poincare
from sympy import symbols, S, binomial, summation, sqrt, cos, sin, atan2, expand_trig,diff,Matrix,Poly
from .disturbing_function import df_coefficient_C,evaluate_df_coefficient_dict,get_df_coefficient_symbol
from scipy.linalg import expm
from .rk_integrator import RKIntegrator, _rk_methods 
from scipy.linalg import expm
from celmech.miscellaneous import getOmegaMatrix, _machine_eps
from collections import defaultdict

from celmech.disturbing_function import list_secular_terms
from celmech.disturbing_function import df_coefficient_C, _add_dicts,_consolidate_dictionary_terms
from celmech.disturbing_function import terms_list_to_HamiltonianCoefficients_dict
from celmech.disturbing_function import resonant_secular_contribution_dictionary
from celmech.poisson_series import DFTermSeries



class LaplaceLagrangeSystem(Poincare):
    r"""
    A class for representing the classical Laplace-Lagrange secular
    solution for a planetary system.
    
    Attributes
    ----------
    eccentricity_matrix : sympy.Matrix
      The matrix :math:`\pmb{S}_e` appearing in the secular equations
      of motion for the eccentricity variables, 

      .. math::
            \frac{d}{dt}(\eta_i + i\kappa_i) = [\pmb{S}_e]_{ij} (\eta_j + i\kappa_j)~.
    
      or, equivalently,

      .. math::
            \frac{d}{dt}\pmb{x} = -i \pmb{S}_e \cdot \pmb{x}

      The matrix is given in symbolic form.
    inclination_matrix : sympy.Matrix
      The matrix :math:`\pmb{S}_I` appearing in the secular equations
      of motion for the eccentricity variables, 

      .. math::
            \frac{d}{dt}(\rho_i + i \sigma_i ) =[\pmb{S}_I]_{ij} (\rho_j + i \sigma_j)~.
    
      or, equivalently,

      .. math::
            \frac{d}{dt}\pmb{y} = -i \pmb{S}_I \cdot \pmb{y}

      The matrix is given in symbolic form.
    Neccentricity_matrix : ndarray
        Numerical value of the eccentricity matrix :math:`\pmb{S}_e`
    Ninclination_matrix : ndarray
        Numerical value of the inclination matrix :math:`\pmb{S}_I`
    Tsec : float
        The secular timescale of the system, defined as the shortest
        secular period among the system's inclination and eccentricity
        modes.
    eccentricity_eigenvalues : ndarray
        Array of the eccentricity mode eigenvalues  (i.e., secular frequencies)
    inclination_eigenvalues : ndarray
        Array of the inclination mode eigenvalues  (i.e., secular frequencies)
    """
    def __init__(self,G,poincareparticles=[]):
        super(LaplaceLagrangeSystem,self).__init__(G,poincareparticles)
        self.params = {S('G'):self.G}
        for i,particle in enumerate(self.particles):
            if i==0:
                continue # skip the star
            m,mu,M,Lambda = symbols('m{0},mu{0},M{0},Lambda{0}'.format(i)) 
            self.params.update({m:particle.m,mu:particle.mu,M:particle.M,Lambda:particle.Lambda})
        self.ecc_entries  = {(j,i):S(0) for i in xrange(1,self.N) for j in xrange(1,i+1)}
        self.inc_entries  = {(j,i):S(0) for i in xrange(1,self.N) for j in xrange(1,i+1)}
        self.tol = np.min([p.m for p in self.particles[1:]]) * np.finfo(np.float).eps
        ps = self.particles[1:]
        self.eta0_vec = np.array([p.eta for p in ps])
        self.kappa0_vec = np.array([p.kappa for p in ps]) 
        self.rho0_vec = np.array([p.rho for p in ps])
        self.sigma0_vec = np.array([p.sigma for p in ps]) 
        self._update()
    @classmethod
    def from_Poincare(cls,pvars):
        """
        Initialize a Laplace-Lagrange system directly from a 
        :class:`celmech.poincare.Poincare` object.
        
        Arguments
        ---------
        pvars : :class:`celmech.poincare.Poincare` 
            Instance of Poincare variables from which to initialize
            Laplace-Lagrange system

        Returns
        -------
        system : :class:`celmech.secular.LaplaceLagrangeSystem`
        """
        return cls(pvars.G,pvars.particles[1:])

    @classmethod
    def from_Simulation(cls,sim):
        """
        Initialize a Laplace-Lagrange system directly from a 
        :class:`rebound.Simulation` object.
        
        Arguments
        ---------
        pvars : :class:`rebound.Simulation` 
            rebound simulation from which to initialize
            Laplace-Lagrange system

        Returns
        -------
        sim : :class:`celmech.secular.LaplaceLagrangeSystem`
        """
        pvars = Poincare.from_Simulation(sim)
        return cls.from_Poincare(pvars)

    @property
    def eccentricity_matrix(self):
        return Matrix([
            [self.ecc_entries[max(i,j),min(i,j)] for i in xrange(1,self.N)]
            for j in xrange(1,self.N) 
            ])

    @property
    def inclination_matrix(self):
        return Matrix([
            [self.inc_entries[max(i,j),min(i,j)] for i in xrange(1,self.N)]
            for j in xrange(1,self.N) 
            ])
    @property 
    def Neccentricity_matrix(self):
        return np.array(self.eccentricity_matrix.subs(self.params)).astype(np.float64)
    @property 
    def Ninclination_matrix(self):
        return np.array(self.inclination_matrix.subs(self.params)).astype(np.float64)
    @property
    def Tsec(self):
        Omega_e = np.max( np.abs(self.eccentricity_eigenvalues()) )
        Omega_i = np.max( np.abs(self.inclination_eigenvalues()) )
        return 2 * np.pi / max(Omega_e,Omega_i)
    def _chop(self,arr):
        arr[np.abs(arr)<self.tol] = 0
        return arr
    def eccentricity_eigenvalues(self):
        return np.linalg.eigvalsh(self.Neccentricity_matrix)
    def inclination_eigenvalues(self):
        answer = np.linalg.eigvalsh(self.Ninclination_matrix)
        return self._chop(answer)

    def secular_solution(self,times,epoch=0):
        """
        Get the solution of the Laplace-Lagrange
        secular equations of motion at the 
        user-specified times.

        Arguments
        ---------
        times : ndarray
            Array of times at which to evaluate 
            the solution to the equations of motion.
        epoch : float, optional
            Current time of system state. Default is 
            t=0.
        Returns
        -------
        soln : dict
            The solution dictionary contains various dynamical
            quantites computed at the input times. 
        """
        e_soln = self.secular_eccentricity_solution(times,epoch)
        solution = {key:val.T for key,val in e_soln.items()}
        T,D = self.diagonalize_inclination()
        R0 = T.T @ self.rho0_vec
        S0 = T.T @ self.sigma0_vec
        t1 = times - epoch
        freqs = np.diag(D)
        cos_vals = np.array([np.cos(freq * t1) for freq in freqs]).T
        sin_vals = np.array([np.sin(freq * t1) for freq in freqs]).T
        S = S0 * cos_vals - R0 * sin_vals    
        R = S0 * sin_vals + R0 * cos_vals
        rho = np.transpose(T @ R.T)
        sigma = np.transpose(T @ S.T)
        Yre = 0.5 * sigma / np.sqrt([p.Lambda for p in self.particles[1:]])
        Yim = -0.5 * rho / np.sqrt([p.Lambda for p in self.particles[1:]])
        kappa,eta = solution['kappa'],solution['eta']
        Xre = kappa / np.sqrt([p.Lambda for p in self.particles[1:]])
        Xim = -eta / np.sqrt([p.Lambda for p in self.particles[1:]])
        Ytozeta = 1 / np.sqrt(1 - 0.5 * (Xre**2 + Xim**2))
        zeta_re = Yre * Ytozeta
        zeta_im = Yim * Ytozeta
        zeta = zeta_re + 1j * zeta_im
        solution.update({
            "rho":rho,
            "sigma":sigma,
            "R":R,
            "S":S,
            "p":zeta_im,
            "q":zeta_re,
            "zeta":zeta,
            "inc":2 * np.arcsin(np.abs(zeta)),
            "Omega":np.angle(zeta)
        })
        return {key:val.T for key,val in solution.items()}

    def secular_eccentricity_solution(self,times,epoch=0):
        T,D = self.diagonalize_eccentricity()
        H0 = T.T @ self.eta0_vec
        K0 = T.T @ self.kappa0_vec
        t1 = times - epoch
        freqs = np.diag(D)
        cos_vals = np.array([np.cos(freq * t1) for freq in freqs]).T
        sin_vals = np.array([np.sin( freq * t1) for freq in freqs]).T
        K = K0 * cos_vals - H0 * sin_vals    
        H = K0 * sin_vals + H0 * cos_vals
        eta = np.transpose(T @ H.T)
        kappa = np.transpose(T @ K.T)
        Xre = kappa / np.sqrt([p.Lambda for p in self.particles[1:]])
        Xim = -eta / np.sqrt([p.Lambda for p in self.particles[1:]])
        Xtoz = np.sqrt(1 - 0.25 * (Xre**2 + Xim**2))
        zre = Xre * Xtoz
        zim = Xim * Xtoz
        solution = {
                "time":times,
                "H":H,
                "K":K,
                "eta":eta,
                "kappa":kappa,
                "k":zre,
                "h":zim,
                "z":zre + 1j * zim,
                "e":np.sqrt(zre*zre + zim*zim),
                "pomega":np.arctan2(zim,zre)
                }
        return {key:val.T for key,val in solution.items()}
    def diagonalize_eccentricity(self):
        r"""
        Solve for matrix S, that diagonalizes the
        matrix T in the equations of motion:

        .. math::

            \frac{d}{dt}(\eta + i\kappa) = i A \cdot (\eta + i\kappa)

        The matrix S satisfies

        .. math::

                T^{T} \cdot A \cdot T = D
        
        where D is a diagonal matrix.
        The equations of motion are decoupled harmonic
        oscillators in the variables (P,Q) defined by 

        .. math::

            H + i K = S^{T} \cdot (\eta + i \kappa)
        
        Returns
        -------
        (T , D) : tuple of n x n numpy arrays
        """
        vals,T = np.linalg.eigh(self.Neccentricity_matrix)
        return T, np.diag(vals)

    def diagonalize_inclination(self):
        r"""
        Solve for matrix U, that diagonalizes the
        matrix B in the equations of motion:
        
        .. math::
            \frac{d}{dt}(\rho + i\sigma) = i B \cdot (\rho + i\sigma)

        The matrix S satisfies

        .. math::
                U^{T} \cdot B \cdot U = D

        where D is a diagonal matrix.
        The equations of motion are decoupled harmonic
        oscillators in the variables :math:`(R,S)` defined by 

        .. math::
            R + i S = U^{T} \cdot (\rho + i \sigma)
        
        Returns
        -------
        (U , D) : tuple of n x n numpy arrays
        """
        vals,U = np.linalg.eigh(self.Ninclination_matrix)
        return U, self._chop(np.diag(vals))
    
    def _update(self):
        G = symbols('G')
        ecc_diag_coeff = df_coefficient_C(*[0 for _ in range(6)],0,0,1,0)
        inc_diag_coeff = df_coefficient_C(*[0 for _ in range(6)],1,0,0,0)
        js_dpomega = 0,0,1,-1,0,0
        js_dOmega = 0,0,0,0,1,-1
        ecc_off_coeff = df_coefficient_C(*js_dpomega,0,0,0,0)
        inc_off_coeff = df_coefficient_C(*js_dOmega,0,0,0,0)
        l1,l2=0,0
        for i in xrange(1,self.N):
            for j in xrange(1,self.N):
                if j==i:
                    continue
                indexIn = min(i,j)
                indexOut = max(i,j)
                particleIn = self.particles[indexIn]
                particleOut = self.particles[indexOut]
                alpha = particleIn.a / particleOut.a
                mIn,muIn,MIn,LambdaIn = symbols('m{0},mu{0},M{0},Lambda{0}'.format(indexIn)) 
                mOut,muOut,MOut,LambdaOut = symbols('m{0},mu{0},M{0},Lambda{0}'.format(indexOut)) 
                Cecc_diag = get_df_coefficient_symbol(*[0 for _ in range(6)],0,0,1,0,l1,l2,indexIn,indexOut)
                Cinc_diag = get_df_coefficient_symbol(*[0 for _ in range(6)],1,0,0,0,l1,l2,indexIn,indexOut)
                aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
                prefactor = -G * mIn * mOut * aOut_inv
                self.params[Cecc_diag] = evaluate_df_coefficient_dict(ecc_diag_coeff,alpha)
                self.params[Cinc_diag] = evaluate_df_coefficient_dict(inc_diag_coeff,alpha)
                if i > j:
                    particleIn = self.particles[indexIn]
                    Cecc = get_df_coefficient_symbol(*js_dpomega,0,0,0,0,l1,l2,indexIn,indexOut)
                    Cinc = get_df_coefficient_symbol(*js_dOmega,0,0,0,0,l1,l2,indexIn,indexOut)
                    alpha = particleIn.a/particleOut.a
                    assert alpha<1, "Particles must be in order by increasing semi-major axis!"
                    Necc_coeff = evaluate_df_coefficient_dict(ecc_off_coeff,alpha)
                    Ninc_coeff = evaluate_df_coefficient_dict(inc_off_coeff,alpha)
                    self.params[Cecc] = Necc_coeff
                    self.params[Cinc] = Ninc_coeff
                    ecc_entry = prefactor  * Cecc / sqrt(LambdaIn) / sqrt(LambdaOut)
                    inc_entry = prefactor  * Cinc / sqrt(LambdaIn) / sqrt(LambdaOut) / 4
                    self.ecc_entries[(indexOut,indexIn)] = ecc_entry
                    self.inc_entries[(indexOut,indexIn)] = inc_entry
                else:
                    pass
                LmbdaI = S('Lambda{}'.format(i))
                self.ecc_entries[(i,i)] += 2 * prefactor * Cecc_diag / LmbdaI
                self.inc_entries[(i,i)] += 2 * prefactor * Cinc_diag / LmbdaI / 4

    def add_first_order_resonance_terms(self,resonances_dictionary):
        for indices,resonance_k in resonances_dictionary.items():
            self.add_first_order_resonance_term(*indices,resonance_k) 

    def add_first_order_resonance_term(self,indexIn, indexOut,jres):
        """
        Include a correction to the Laplace-Lagrange 
        secular equations of motion that arise due to a nearby 
        first-order mean motion resonance between two planets.

        Note-- corrections are not valid for planets in resonance!
    
        Arguments
        ---------
        indexIn : int
            Index of inner planet near resonance.
        indexOut : int
            Index of inner planet near resonance.
        jres : int
            Specify the jres:jres-1 mean motion resonance.
        """
        assert indexIn < indexOut, "Input 'indexIn' must be less than 'indexOut'."
        particleIn = self.particles[indexIn]
        particleOut = self.particles[indexOut]
        alpha = particleIn.a / particleOut.a
        G = symbols('G')
        mIn,muIn,MIn,LambdaIn = symbols('m{0},mu{0},M{0},Lambda{0}'.format(indexIn)) 
        mOut,muOut,MOut,LambdaOut = symbols('m{0},mu{0},M{0},Lambda{0}'.format(indexOut)) 
        l1,l2 = 0,0
        CIn = get_df_coefficient_symbol(*[jres,1-jres,-1,0,0,0],0,0,0,0,l1,l2,indexIn,indexOut)
        COut = get_df_coefficient_symbol(*[jres,1-jres,0,-1,0,0],0,0,0,0,l1,l2,indexIn,indexOut)
        self.params[CIn] = evaluate_df_coefficient_dict(df_coefficient_C(*[jres,1-jres,-1,0,0,0],0,0,0,0),alpha)
        self.params[COut] = evaluate_df_coefficient_dict(df_coefficient_C(*[jres,1-jres,0,-1,0,0],0,0,0,0),alpha)
        aOut_inv = G*MOut*muOut*muOut / LambdaOut / LambdaOut  
        eps = -G * mIn * mOut * aOut_inv
        omegaIn = G * G * MIn * MIn * muIn**3 / (LambdaIn**3)
        omegaOut = G * G * MOut * MOut * muOut **3 / (LambdaOut**3)
        domegaIn = -3 * G * G * MIn * MIn * muIn**3 / (LambdaIn**4)
        domegaOut = -3 * G * G * MOut * MOut * muOut**3 / (LambdaOut**4)
        kIn = 1 - jres
        kOut = jres
        k_Domega_k = kIn**2 * domegaIn + kOut**2 * domegaOut
        prefactor = (k_Domega_k / (kIn * omegaIn + kOut * omegaOut)**2) 
        xToXIn = sqrt(2/LambdaIn)
        xToXOut = sqrt(2/LambdaOut)

        InIn = eps**2 * prefactor * CIn * CIn * xToXIn * xToXIn / 4
        InOut = eps**2 * prefactor * COut * CIn * xToXOut * xToXIn / 4
        OutOut = eps**2 * prefactor * COut * COut * xToXOut * xToXOut / 4
        
        self.ecc_entries[(indexIn,indexIn)] += InIn
        # Note-- only upper entries are stored so 
        # changing (indexOut,indexIn) also implicitly 
        # changes (indexIn,indexOut)
        self.ecc_entries[(indexOut,indexIn)] += InOut
        self.ecc_entries[(indexOut,indexOut)] += OutOut
        
def _get_pair_SecularHamiltonian_coefficients(Nmin,Nmax,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out,res_jk_list=[]):
    """
    Calculate the coefficients appearing in secular Hamiltonian expansion.
    """
    terms = list_secular_terms(Nmin,Nmax)
    extra_args = G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out
    dsec = terms_list_to_HamiltonianCoefficients_dict(terms,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out)
    for j,k in res_jk_list:
        dres = resonant_secular_contribution_dictionary(j,k,Nmin,Nmax,*extra_args)
        dsec = _add_dicts(dsec,dres)
        dsec = _consolidate_dictionary_terms(dsec)
    return _consolidate_dictionary_terms(dsec)
    
class _SecularDerivativesSystem():
    """
    A class for calculating the derivatives of a secular system.
    """
    def __init__(self,N,Hamiltonian_coefficients_dictionary,Lambda0):

        self.Lambda0 = np.array(Lambda0[1:])
        self.rtLambda0_inv = 1 / np.sqrt(Lambda0[1:])
        self.qp_to_XY_factors = np.concatenate((self.rtLambda0_inv,0.5*self.rtLambda0_inv))
        self.N=N
        self.Npl = self.N-1
        self.Ndim = 4*self.Npl

        self.DFSeries_dict=dict()
        for key,term_coeff_dict in Hamiltonian_coefficients_dictionary.items():
            i,j=key
            Lambda0In = self.Lambda0[i-1]
            Lambda0Out = self.Lambda0[j-1]
            if len(term_coeff_dict) > 0:
                self.DFSeries_dict[key] = DFTermSeries(term_coeff_dict,Lambda0In,Lambda0Out)

    def Hamiltonian_from_qp_vec(self,qp_vec):
        """
        Compute Hamiltonian of the operator from the
        equations of motion for the 'qp_vec' variables
        returned by method 'state_vec_to_qp_vec'.

        Arguments
        ---------
        qp_vec : ndarray
          Input variable vector in the from
           [eta1,eta2,...,etaN,rho1,...,rhoN,kappa1,...,kappaN,sigma1,...sigmaN]

        Returns
        -------
        Hamiltonian : float
          The value of the Hamiltonian (i.e., the sum of the disturbing
          function terms modeled by the operator)
        """
        l = np.zeros(2)
        eta,rho,kappa,sigma = qp_vec.reshape(-1,self.Npl)
        H = eta * self.rtLambda0_inv
        K = kappa * self.rtLambda0_inv
        R = 0.5 * rho * self.rtLambda0_inv
        S = 0.5 * sigma * self.rtLambda0_inv
        Hamiltonian = 0.
        for iPair,series in self.DFSeries_dict.items():
            iIn, iOut = iPair
            indices = np.array(iPair) - 1
            X = K[indices] - 1j * H[indices]
            Y = S[indices] - 1j * R[indices]
            XYvec = np.concatenate((X,Y))
            dH = series._evaluate(l,XYvec)
            Hamiltonian += dH
        return Hamiltonian

    def deriv_from_qp_vec(self,qp_vec):
        """
        Compute the time derivatives from the
        equations of motion for the 'qp_vec'
        variables returned by method 'state_vec_to_qp_vec'.

        Arguments
        ---------
        qp_vec : ndarray
          Input variable vector in the from
           [eta1,eta2,...,etaN,rho1,...,rhoN,kappa1,...,kappaN,sigma1,...sigmaN]

        Returns
        -------
        qp_vec_dot : ndarray
          Time derivative of qp_vec.
        """
        derivs = np.zeros(self.Ndim)
        l = np.zeros(2)
        eta,rho,kappa,sigma = qp_vec.reshape(-1,self.Npl)
        H = eta * self.rtLambda0_inv
        K = kappa * self.rtLambda0_inv
        R = 0.5 * rho * self.rtLambda0_inv
        S = 0.5 * sigma * self.rtLambda0_inv
        for iPair,series in self.DFSeries_dict.items():
            iIn, iOut = iPair
            indices = np.array(iPair) - 1
            X = K[indices] - 1j * H[indices]
            Y = S[indices] - 1j * R[indices]
            XYvec = np.concatenate((X,Y))
            _, _deriv= series._evaluate_with_derivs(l,XYvec)
            index_list = np.array([
                iIn,iOut,
                iIn + self.Npl,iOut + self.Npl,
                iIn + 2*self.Npl,iOut + 2*self.Npl,
                iIn + 3*self.Npl,iOut + 3*self.Npl
                ]) - 1
            for i,I in enumerate(index_list):
                derivs[I] += _deriv[i]
        return derivs

    def deriv_and_jacobian_from_qp_vec(self,qp_vec):
        """
        Compute the time derivatives and Jacobian from the
        equations of motion for the 'qp_vec' variables returned
        by method 'state_vec_to_qp_vec'.

        Arguments
        ---------
        qp_vec : ndarray shape (4 * Nplanet,)
          Input variable vector in the from
           [eta1,eta2,...,etaN,rho1,...,rhoN,kappa1,...,kappaN,sigma1,...sigmaN]

        Returns
        -------
        qp_vec_dot : ndarray, shape (4 * Nplanet,)
          Time derivative of qp_vec.

        qp_vec_dot_jac : ndarray, shape (4 * Nplanet, 4 * Nplanet)
          Jacobian matrix of the equations of motion for the
          variables contained in qp_vec.
        """
        derivs = np.zeros(self.Ndim)
        jac = np.zeros((self.Ndim,self.Ndim))
        l = np.zeros(2)
        eta,rho,kappa,sigma = qp_vec.reshape(-1,self.Npl)
        H = eta * self.rtLambda0_inv
        K = kappa * self.rtLambda0_inv
        R = 0.5 * rho * self.rtLambda0_inv
        S = 0.5 * sigma * self.rtLambda0_inv
        for iPair,series in self.DFSeries_dict.items():
            iIn, iOut = iPair
            indices = np.array(iPair) - 1
            X = K[indices] - 1j * H[indices]
            Y = S[indices] - 1j * R[indices]
            XYvec = np.concatenate((X,Y))
            _, _deriv, _jac, _ = series._evaluate_with_jacobian(l,XYvec)
            index_list = np.array([
                iIn,iOut,
                iIn + self.Npl,iOut + self.Npl,
                iIn + 2*self.Npl,iOut + 2*self.Npl,
                iIn + 3*self.Npl,iOut + 3*self.Npl
                ]) - 1
            for i,I in enumerate(index_list):
                derivs[I] += _deriv[i]
                for j,J in enumerate(index_list):
                    jac[I,J] += _jac[i,j]
        return derivs, jac

class SecularRKIntegrator(RKIntegrator):
    """
    Integrator for direct integration of secular equations of motion.
    """
    def __init__(self,N,hamiltonian_coefficients_dictionary,Lambda0s,dt,
                 rtol=_machine_eps,
                 atol=0,
                 rk_method='ImplicitMidpoint',
                 rk_root_method='Newton',
                 max_iter=10
                ):
        self._derivatives = _SecularDerivativesSystem(N,hamiltonian_coefficients_dictionary,Lambda0s)
        Ndim = 4 * (N-1)
        super(SecularRKIntegrator,self).__init__(
            self._derivatives.deriv_from_qp_vec,
            self._derivatives.deriv_and_jacobian_from_qp_vec,
            Ndim, dt, rtol, atol, rk_method, rk_root_method, max_iter
        )
    
        self.step = self.rk_step
        
    def init_step(self,qpvec):
        return qpvec
    
    def final_step(self,qpvec):
        return qpvec
    
    def calculate_energy(self,qp):
        return self._derivatives.Hamiltonian_from_qp_vec(qp)

class SecularSplittingIntegrator(RKIntegrator):
    """
    Integrator for applying splitting method to secular equations of motion.
    """
    def __init__(self,N,hamiltonian_coefficients_dictionary,Lambda0s,dt,
                 rtol=_machine_eps,
                 atol=0,
                 rk_method='ImplicitMidpoint',
                 rk_root_method='Newton',
                 max_iter=10
                ):
        
        all_hcoeffs = hamiltonian_coefficients_dictionary.copy()
        Npl = N-1
        Ndim = 4 * Npl
        self.Se = np.zeros((Npl,Npl))
        self.SI = np.zeros((Npl,Npl))
        for i in range(Npl):
            for j in range(i+1,Npl):
                hcoeffs=all_hcoeffs[(i+1,j+1)]
                # Get second-order eccentricity terms and add them to Laplace-Lagrange matrix
                self.Se[i,i] += hcoeffs.pop(((0,0,0,0,0,0),(0,0,1,0)),0)
                self.Se[j,j] += hcoeffs.pop(((0,0,0,0,0,0),(0,0,0,1)),0)
                term = hcoeffs.pop(((0,0,1,-1,0,0),(0,0,0,0)),0)
                self.Se[i,j] += 0.5 * term
                self.Se[j,i] += 0.5 * term
                # Get second-order inclination terms and add them to Laplace-Lagrange matrix
                self.SI[i,i] += hcoeffs.pop(((0,0,0,0,0,0),(1,0,0,0)),0)
                self.SI[j,j] += hcoeffs.pop(((0,0,0,0,0,0),(0,1,0,0)),0)
                term = hcoeffs.pop(((0,0,0,0,1,-1),(0,0,0,0)),0)
                self.SI[i,j] += 0.5 * term
                self.SI[j,i] += 0.5 * term
        rtLambda0s_inv = 1/np.sqrt(Lambda0s[1:])
        
        self.Se = 2 * np.diag(rtLambda0s_inv) @ self.Se @ np.diag(rtLambda0s_inv)
        self.SI = 0.5 * np.diag(rtLambda0s_inv) @ self.SI @ np.diag(rtLambda0s_inv)                
        self._derivatives = _SecularDerivativesSystem(N,all_hcoeffs,Lambda0s)
        Se,SI = self.Se,self.SI
        Zeros = np.zeros((Npl,Npl))
        self.generator_matrix = np.block([
            [Zeros,Zeros,Se,Zeros],
            [Zeros,Zeros,Zeros,SI],
            [-Se,Zeros,Zeros,Zeros],
            [Zeros,-SI,Zeros,Zeros],
        ])
        self._update_A_step_matrices(dt)
        self.corrector= False
        
        super(SecularSplittingIntegrator,self).__init__(
            self._derivatives.deriv_from_qp_vec,
            self._derivatives.deriv_and_jacobian_from_qp_vec,
            Ndim, dt, rtol, atol, rk_method, rk_root_method, max_iter
        )
        
        self.Bstep = self.rk_step
    
    def _update_A_step_matrices(self,dt):
        gen_matrix = self.generator_matrix
        self.A_half_step_forward_matrix = expm(0.5 * dt * gen_matrix)
        self.A_half_step_backward_matrix = expm(-0.5 * dt * gen_matrix)
        self.A_full_step_forward_matrix = self.A_half_step_forward_matrix @ self.A_half_step_forward_matrix

  
    @property
    def dt(self):
        return self._dt
    
    @dt.setter
    def dt(self,dt):
        self._update_A_step_matrices(dt)
        self._dt = dt
        
    def init_step(self,qpvec):
        if self.corrector:
                qpvec = self.corrector3(qpvec,self.dt)
        return self.A_half_step_forward_matrix @ qpvec
    
    def step(self,qpvec):
        qpvec = self.Bstep(qpvec)
        return self.A_full_step_forward_matrix @ qpvec
    
    def final_step(self,qpvec):
        qpvec = self.A_half_step_backward_matrix @ qpvec
        if self.corrector:
            qpvec = self.corrector3inv(qpvec,self.dt)
        return qpvec
    
    def _apply_A_step_for_dt(self,qpvec,h):
        return expm(h * self.generator_matrix) @ qpvec
    
    def _apply_B_step_for_dt(self,qpvec,h):
        h0 = self.dt
        self.dt = h
        qpvec1 = self.rk_step(qpvec)
        self.dt = h0
        return qpvec1
        
    def X(self, qpvec, a, b, h):
        qpvec = self._apply_A_step_for_dt(qpvec,-a*h)
        qpvec = self._apply_B_step_for_dt(qpvec,b*h)
        qpvec = self._apply_A_step_for_dt(qpvec,a*h)
        return qpvec

    def Z(self, qpvec, a, b, h):
        qpvec = self.X(qpvec, -a, -b, h)
        qpvec = self.X(qpvec, a, b, h)
        return qpvec

    def corrector3(self, qpvec, h):
        alpha = (7./40.)**0.5
        beta = 1/48./alpha
        a1 = -alpha
        a2 = alpha
        b2 = beta/2.
        b1 = -beta/2.
        
        qpvec = self.Z(qpvec, a2, b2, h)
        qpvec = self.Z(qpvec, a1, b1, h)
        return qpvec

    def corrector3inv(self, qpvec, h):
        alpha = (7./40.)**0.5
        beta = 1/48./alpha
        a1 = -alpha
        a2 = alpha
        b2 = beta/2.
        b1 = -beta/2.
        
        qpvec = self.Z(qpvec, a1, -b1, h)
        qpvec = self.Z(qpvec, a2, -b2, h)
        return qpvec

    def calculate_energy(self,qp):
        Epert = self._derivatives.Hamiltonian_from_qp_vec(qp)
        Omega = getOmegaMatrix(self.Ndim//2)
        E0 = -0.5 * qp @ Omega @ self.generator_matrix @ qp
        return E0 + Epert




class SecularSystemSimulation():
    """
    A class for integrating the secular equations of motion governing a planetary system.

    Arguments
    ---------
    state : :class:`celmech.poincare.Poincare`
        The initial dynamical state of the system.

    dt : float, optional
        The timestep to use for the integration. Either dt or dtFraction must be
        specified.

    dtFraction : float, optional
        Set the timestep to a constant fraction the period of shortest-period linear
        secular eigenmode.

    max_order : int, optional
        The maximum order of disturbing function terms to include in the integration.
        By default, the equations of motion include terms up to 4th order.

    method : str
        Integration method to use. Options include:
            - 'RK' - Direct integration of equations of motion using an Runge-Kutta integration method.
            - 'splitting' - Applies splitting to the flow generated by the system's Hamiltonian.

    resonances_to_include : dict, optional
        A dictionary containing information that sets the list of MMRs for which the
        secular contribution will be accounted for (at second order on planet masses).
        Dictionary key and value pairs are specified in the form

              .. code::

                {(iIn,iOut) : [(j_0,k_0),...(j_N,k_N)]}

        include the resonances :math:`j_i` : :math:`j_i-k_i` with :math:`i=0,...,N` between planets
        iIn and iOut. Note that harmonics should **NOT** be explicitly included. I.e.,
        if (2,1) appears in the list [(j_0,k_0),...(j_N,k_N)] then the term (4,2) should
        **NOT** also appear; these terms' contributions will be added automatically when
        the (2,1) term is added.
        By default, no MMRs are included.

    rk_kwargs : dict, optional
        Keyword arguments that determine details of the Runge-Kutta scheme used to integrate
        equations of motion. Keywoards include:
            - :code:`rtol`: float
                Relative tolerance of root-finding step in the implicit Runge-Kutta scheme.
                Default is set to machine precision.

            - :code:`atol`: float
                Absolute tolerance of the root-finding step in the implicit Runge-Kutta scheme.
                Default is 0 so that tolerance is set solely by :code:`rtol` argument.

            - :code:`rk_method`: str
                Runge-Kutta method to use.  Available options include:
                    - 'ImplicitMidpoint'
                    - 'LobattoIIIB'
                    - 'GL4'
                    - 'GL6'
                    - 'ExplicitMidpoint'
                    - 'RK4'
                'GL4' and 'GL6' are `Gauss-Legendre methods <https://en.wikipedia.org/wiki/Gauss–Legendre_method>`_ of order 4 and 6, respectively.
                'ImplicitMidpoint', 'GL4', and 'GL6' are symplectic methods while 'LobattoIIIB' is a 4th order time-reversible method (but not symplectic).
                'ExplicitMidpoint' and 'RK4' are explicit methods that are neither symplectic nor time-reversible.

            - :code:`rk_root_method` : str
                Method to use for root-finding during implicit RK step. Available options are:
                        - 'Newton'
                        - 'qausi-Newton'
                        - 'fixed_point'
                        - 'explicit'
                'Newton' (default) uses Newton's method whereas 'fixed_point' uses a fixed point iteration method.
                Newton's method requires computing the Jacobian of the equations of motion but has quadratic convergence.
    """

    def __init__(self,
                 state,
                 dt=None,dtFraction=None,
                 max_order=4,method='RK',
                 resonances_to_include={},
                 rk_kwargs = {}
               ):

        self.state = state
        self._hamiltonian_coefficients_dictionary = dict()
        Npl = self.N-1

        self.resonances_to_include = resonances_to_include
        self._hamiltonian_coefficients_dictionary = {(i+1,j+1):{} for i in range(Npl) for j in range(i+1,Npl)}
        self._max_order = 0
        self._update_matrcies_and_coefficient_dictionary(max_order)

        ecc_eigenvalues = np.linalg.eigvals(self.Se)
        inc_eigenvalues = np.linalg.eigvals(self.SI)
        self.Tsec_e = np.abs(2 * np.pi / ecc_eigenvalues)
        self.Tsec_inc = np.abs(2 * np.pi / inc_eigenvalues)
        self.Tsec = min(np.min(self.Tsec_e),np.min(self.Tsec_inc))
        self.t = 0

        if dt:
            self._dt = dt
        elif dtFraction:
            self._dt = dtFraction * self.Tsec
        else:
            raise AttributeError("Must specify either 'dt' or 'dtFraction'")

        self.secular_rk_integrator = SecularRKIntegrator(
            self.N,
            self.Hamiltonian_coefficients_dictionary,
            self.Lambda0s,
            self.dt,
            **rk_kwargs
        )

        self.secular_splitting_integrator = SecularSplittingIntegrator(
            self.N,
            self.Hamiltonian_coefficients_dictionary,
            self.Lambda0s,
            self.dt,
            **rk_kwargs
        )

        self.method = method

    @property
    def method(self):
        """
        Integration method to use. 
        
        Valid options are
         - RK
         - splitting
        """
        return self._method_name

    @method.setter
    def method(self,method):
        if method == 'RK':
            self._method_name = 'RK'
            self._integrator = self.secular_rk_integrator
        elif method == 'splitting':
            self._method_name = 'splitting'
            self._integrator = self.secular_splitting_integrator
        else:
            raise ValueError("{} is not a valid method option.".format(method))

    @property
    def t(self):
        """Simulation time"""
        return self.state.t
    @t.setter
    def t(self,val):
        self.state.t = val

    @property
    def dt(self):
        """Simulation time step"""
        return self._dt

    @dt.setter
    def dt(self,h):
        self._dt = h
        self.secular_rk_integrator.dt = h
        self.secular_splitting_integrator.dt = h

    @property
    def Lambda0s(self):
        """Values of particle's :math:`\Lambda` canonical momenta"""
        return np.array([p.Lambda for p in self.state.particles])

    @property
    def state_vector(self):
        """Current state vector of systems' canonical Poincare variables."""
        state_vec = []
        for p in self.state.particles[1:]:
            state_vec += [p.kappa,p.eta,p.Lambda,p.l,p.sigma,p.rho]
        return np.array(state_vec)

    @property
    def max_order(self):
        """Maximum order of disturbing function expansion in :math:`e` and :math:`I`"""
        return self._max_order

    @max_order.setter
    def max_order(self,new_max_order):
        self._update_matrcies_and_coefficient_dictionaries(new_max_order)
        self._set_up_derivative_systems()
        self.EccentricityLinearEvolutionOperatorMatrix = expm(0.5 * self.dt * self.Se)
        self.InclinationLinearEvolutionOperatorMatrix = expm(0.5 * self.dt * self.SI)

    @property
    def N(self):
        """Number of particles (including central body)"""
        return self.state.N

    @property
    def Npl(self):
        return self.N - 1

    @property
    def G(self):
        """Value of the gravitational constant"""
        return self.state.G

    @property
    def Hamiltonian_coefficients_dictionary(self):
        """
        A dictionary containing the coefficients appearing in the Hamiltonian.
        Dictionary is organized by pair-wise interaction terms.
        """
        return self._hamiltonian_coefficients_dictionary

    @classmethod
    def from_Simulation(cls,sim,
                        dt=None,dtFraction=None,
                        max_order=4,method='RK',
                        resonances_to_include={},
                        rk_kwargs = {}
                       ):
        """
        Initialize a :class:`SecularSystemSimulation <celmech.secular.SecularSystemSimulation>` object
        from a rebound simulation.

        Arguments
        ---------
        sim : :class:`rebound.Simulation`
            REBOUND simulation to convert to :class:`SecularSystemSimulation <celmech.secular.SecularSystemSimulation>`

        dt : float, optional
            The timestep to use for the integration. Either dt or dtFraction must be
            specified.

        dtFraction : float, optional
            Set the timestep to a constant fraction the period of shortest-period linear
            secular eigenmode.

        max_order : int, optional
            The maximum order of disturbing function terms to include in the integration.
            By default, the equations of motion include terms up to 4th order.

        method : str
            Integration method to use. Options include:
            - 'RK' - Direct integration of equations of motion using an
              Runge-Kutta integration method.
            - 'splitting' - Applies splitting to the flow generated by the
              system's Hamiltonian.

        resonances_to_include : dict, optional
            A dictionary containing information that sets the list of MMRs for which the
            secular contribution will be accounted for (at second order on planet masses).
            Dictionary key and value pairs are specified in the form

                  .. code::

                    {(iIn,iOut) : [(j_0,k_0),...(j_N,k_N)]}

            include the resonances :math:`j_i` : :math:`j_i-k_i` with :math:`i=0,...,N` between planets
            iIn and iOut. Note that harmonics should **NOT** be explicitly included. I.e.,
            if (2,1) appears in the list [(j_0,k_0),...(j_N,k_N)] then the term (4,2) should
            **NOT** also appear; these terms' contributions will be added automatically when
            the (2,1) term is added.
            By default, no MMRs are included.

        rk_kwargs : dict, optional
            Keyword arguments that determine details of the Runge-Kutta scheme used to integrate
            equations of motion. Keywoards include:
                - :code:`rtol`: float
                    Relative tolerance of root-finding step in the implicit Runge-Kutta scheme.
                    Default is set to machine precision.

                - :code:`atol`: float
                    Absolute tolerance of the root-finding step in the implicit Runge-Kutta scheme.
                    Default is 0 so that tolerance is set solely by :code:`rtol` argument.

                - :code:`rk_method`: str
                    Runge-Kutta method to use.  Available options include:
                        - 'ImplicitMidpoint'
                        - 'LobattoIIIB'
                        - 'GL4'
                        - 'GL6'
                    'GL4' and 'GL6' are `Gauss-Legendre methods <https://en.wikipedia.org/wiki/Gauss–Legendre_method>`_ of order 4 and 6, respectively.
                    'ImplicitMidpoint', 'GL4', and 'GL6' are symplectic methods while 'LobattoIIIB' is a 4th order time-reversible method (but not symplectic).

                - :code:`rk_root_method` : str
                    Method to use for root-finding during implicit RK step. Available options are:
                            - 'Newton'
                            - 'qausi-Newton'
                            - 'fixed_point'
                            - 'explicit'
                    'Newton' (default) uses Newton's method whereas 'fixed_point' uses a fixed point iteration method.
                    Newton's method requires computing the Jacobian of the equations of motion but has quadratic convergence.
    """


        pvars = Poincare.from_Simulation(sim)
        return cls(pvars,dt,dtFraction,max_order,method,resonances_to_include,rk_kwargs)



    def update_state_from_vector(self,state_vec):
        vecs =  np.reshape(state_vec,(-1,6))
        for vals,p in zip(vecs,self.state.particles[1:]):
            p.kappa,p.eta,p.Lambda,p.l,p.sigma,p.rho = vals


    def _update_matrcies_and_coefficient_dictionary(self,new_max_order):

        ps=self.state.particles
        G = self.G
        Npl = self.Npl
        old_max_order = self._max_order
        self.Se = np.zeros((Npl,Npl))
        self.SI = np.zeros((Npl,Npl))

        for i in range(1,self.N):
            I = i-1
            for j in range(i+1,self.N):
                J = j-1

                pIn = ps[i]
                pOut = ps[j]
                extra_args = G,pIn.m,pOut.m,pIn.M,pOut.M,pIn.Lambda,pOut.Lambda
                res_jk_list = self.resonances_to_include.get((i,j),[])
                hamiltonian_coeff_dict = self._hamiltonian_coefficients_dictionary[(i,j)]

                if new_max_order > self.max_order:

                    coeff_dict = _get_pair_SecularHamiltonian_coefficients(
                        old_max_order+2,
                        new_max_order,
                        *extra_args,
                        res_jk_list
                    )

                    hamiltonian_coeff_dict.update(coeff_dict)

                elif new_max_order < self.max_order:
                    for k,z in hamiltonian_coeff_dict:
                        order= np.sum(np.abs(k)) + 2 * np.sum(z)
                        if order > new_max_order:
                            hamiltonian_coeff_dict.pop((k,z))

                # Get second-order eccentricity terms and add them to Laplace-Lagrange matrix
                self.Se[I,I] += hamiltonian_coeff_dict.get(((0,0,0,0,0,0),(0,0,1,0)),0)
                self.Se[J,J] += hamiltonian_coeff_dict.get(((0,0,0,0,0,0),(0,0,0,1)),0)
                term = hamiltonian_coeff_dict.get(((0,0,1,-1,0,0),(0,0,0,0)),0)
                self.Se[I,J] += 0.5 * term
                self.Se[J,I] += 0.5 * term

                # Get second-order inclination terms and add them to Laplace-Lagrange matrix
                self.SI[I,I] += hamiltonian_coeff_dict.get(((0,0,0,0,0,0),(1,0,0,0)),0)
                self.SI[J,J] += hamiltonian_coeff_dict.get(((0,0,0,0,0,0),(0,1,0,0)),0)
                term = hamiltonian_coeff_dict.get(((0,0,0,0,1,-1),(0,0,0,0)),0)
                self.SI[I,J] += 0.5 * term
                self.SI[J,I] += 0.5 * term

        Lambda0s = self.Lambda0s[1:]
        rtLambda0s_inv = 1/np.sqrt(Lambda0s)
        self.Se = 2 * np.diag(rtLambda0s_inv) @ self.Se @ np.diag(rtLambda0s_inv)
        self.SI = 0.5 * np.diag(rtLambda0s_inv) @ self.SI @ np.diag(rtLambda0s_inv)

    def state_to_qp_vec(self):
        r"""
        Convert full state vector to vector of variables
        that enter in the secular equations.

        Returns
        -------
        qp_vec : ndarray
         Vector containing eccentricity and inclination
         variables :math:`\eta,\rho,\kappa,\sigma`. The variables
         are returned in the order:

         .. math::

            (\eta_1,\eta_2,...,\eta_N,\rho_1,...,\rho_N,\kappa_1,...,\kappa_N,\sigma_1,...,\sigma_N)

        """
        vecs = np.reshape(self.state_vector,(-1,6))
        kappa = vecs[:,0]
        eta = vecs[:,1]
        sigma = vecs[:,4]
        rho = vecs[:,5]
        return np.concatenate((eta,rho,kappa,sigma))


    def integrate(self,time,exact_finish_time=False):
        assert time >= self.t, "Backward integration is currently not implemented."
        Nstep = int( np.ceil( (time-self.t) / self.dt) )
        Npl=self.Npl
        state_vec = self.state_vector
        qp = self.state_to_qp_vec()

        qp = self._integrator.init_step(qp)
        for _ in xrange(Nstep):
            qp = self._integrator.step(qp)
        qp = self._integrator.final_step(qp)

        if exact_finish_time:
            warnings.warn("Exact finish time is not currently implemented.")

        for i in xrange(Npl):
            eta,kappa,rho,sigma = symbols("eta{0},kappa{0},rho{0},sigma{0}".format(i+1))
            # eta
            self.state.qp[eta] =  qp[i]
            # state_vec[6*i+1] = qp[i]
            # kappa
            self.state.qp[kappa] = qp[i + 2 * Npl] 
            #state_vec[6*i] = qp[i + 2 * Npl]
            # rho
            #state_vec[6*i+5] = qp[i + Npl]
            self.state.qp[rho] = qp[i +  Npl] 
            # sigma
            #state_vec[6*i+4] = qp[i + 3 * Npl]
            self.state.qp[sigma] = qp[i + 3 * Npl] 
        #self.update_state_from_vector(state_vec)
        self.state.t += Nstep * self.dt


    def calculate_energy(self):
        qp = self.state_to_qp_vec()
        return self._integrator.calculate_energy(qp)

    def calculate_AMD(self):
        return sum([p.Q + p.Gamma for p in self.state.particles[1:]])

    def _get_symbol_vec(self,sybmol_str):
        return symbols(",".join([r'{}_{}'.format(sybmol_str,i) for i in range(1,self.Npl+1)]))

    def diagonalizing_tranformations(self):
        r"""
        Calculate transformations 

        .. math::
            \begin{align}
                D_e &=& T_e^\mathrm{T}S_e T_e\\
                D_I &=& T_I^\mathrm{T}S_I T_I\\

        that diagonalize the matrices :math:`S_e` 
        and :math:`S_I`.

        Returns
        -------
        Te : ndarray
        TI : ndarray
        De : ndarray
        DI : ndarray
        """
        vals_e,Te = np.linalg.eigh(self.Se)
        vals_I,TI = np.linalg.eigh(self.SI)
        vals_I[0] = 0
        De = np.diag(vals_e)
        DI = np.diag(vals_I)
        return Te,TI,De,DI

    def Hamiltonian_as_polynomial(self, transformed = False):
        r"""
        Return the Hamiltonian of the secular system as a :py:obj:`sympy.Poly`
        object.

        Argmunets
        ---------
        transformed : bool, optional
            If :code:`True`, transform to the complex canonical variables
            :math:`\pmb{u}` and :math:`\pmb{v}`, that represent the proper
            secular modes and are related to the usual complex canonical
            variables via :math:`\pmb{x} = T_e\cdot\pmb{u}` and
            :math:`\pmb{y}=T_I\cdot\pmb{v}`.

        Returns
        -------
        H : sympy.Poly
            Polynomial representation of the Hamiltonian.
        """
        H = self._hamiltonian_to_poly(not transformed)
        if transformed:
            Npl = self.Npl
            get_symbol_vec = self._get_symbol_vec
            u=get_symbol_vec("u")
            ubar=get_symbol_vec(r"\bar{u}")
            v=get_symbol_vec("v")
            vbar=get_symbol_vec(r"\bar{v}")
            Te,TI,De,DI = self.diagonalizing_tranformations()
            cvars = H.gens
            transformed_vars = np.concatenate((Te.dot(u),TI.dot(v),Te.dot(ubar),TI.dot(vbar)))
            rule={a:b for a,b in zip(cvars,transformed_vars)}
            H = H.as_expr().subs(rule)
            H += De.dot(u).dot(ubar) + DI.dot(v).dot(vbar)
            H = H.as_poly(u+v+ubar+vbar)
        return H

    def _hamiltonian_to_poly(self,include_linear_terms):
        get_symbol_vec = self._get_symbol_vec
        Npl = self.Npl
        x=get_symbol_vec("x")
        xbar=get_symbol_vec(r"\bar{x}")
        y=get_symbol_vec("y")
        ybar=get_symbol_vec(r"\bar{y}")
        poly_dict = defaultdict(float)
        hdict = self.Hamiltonian_coefficients_dictionary
        Lambdas = self.Lambda0s
        for ij,terms_dict in hdict.items():
            i,j=ij
            Lin,Lout = Lambdas[i],Lambdas[j]
            coeff_dict = _XYcoeff_to_xycoeff(terms_dict,Lin,Lout)
            for kz, val in coeff_dict.items():
                pows = _ijkz_to_Vpows(i,j,*kz,Npl)
                cpows = _conj_powers(pows)
                poly_dict[tuple(pows)]+=0.5*val
                poly_dict[tuple(cpows)]+=0.5*val
        H = Poly.from_dict(poly_dict,x+y+xbar+ybar)
        if include_linear_terms:
            H2 = np.array(xbar).dot(self.Se.dot(x))
            H2 += np.array(ybar).dot(self.SI.dot(y))
            H += H2.as_poly()
        return H

def _ijkz_to_Vpows(ii,jj,k,z,N):
    # Planets indexed from 1
    # array entries indexed from 0
    i=ii-1
    j=jj-1
    vpows = np.zeros(4*N,dtype=int)
    k1,k2,k3,k4,k5,k6 = k
    z1,z2,z3,z4 = z
    if k3 != 0:
        vpows[i + N * (1 - np.sign(k3))] = abs(k3)
    if k4 != 0:
        vpows[j + N * (1 - np.sign(k4))] = abs(k4)
    if k5 != 0:
        vpows[i + N + N * (1 - np.sign(k5))] = abs(k5)
    if k6 != 0:
        vpows[j + N + N * (1 - np.sign(k6))] = abs(k6)

    # eccentricity terms
    vpows[i] += z3
    vpows[i + 2 * N] += z3
    vpows[j] += z4
    vpows[j + 2 * N] += z4

    # inclination terms
    vpows[i + N] += z1
    vpows[i + 3 * N] += z1
    vpows[j + N] += z2
    vpows[j + 3 * N] += z2
    
    return vpows

def _conj_powers(powers_arr):
    N = powers_arr.shape[0]//2
    return np.concatenate((powers_arr[N:],powers_arr[:N]))

def _XYcoeff_to_xycoeff(d,Lin,Lout):
    dnew = dict()
    for kz,val in d.items():
        k,z=kz
        Xpow = abs(k[2]) + 2 * z[2]
        X1pow = abs(k[3]) + 2 * z[3]
        Ypow = abs(k[4]) + 2 * z[0]
        Y1pow = abs(k[5]) + 2 * z[1]
        factor = (2/Lin)**(Xpow/2) * (2/Lout)**(X1pow/2) * (0.5/Lin)**(Ypow/2) * (0.5/Lout)**(Y1pow/2)
        dnew[kz] = val * factor
    return dnew





from sympy import S, diff, lambdify, symbols, sqrt, cos,sin, numbered_symbols, simplify,binomial, hyper, hyperexpand, Function, factorial,elliptic_k,elliptic_e, expand_trig, Function,bell
from sympy import I,exp,series
from . import clibcelmech
from ctypes import Structure, c_double, POINTER, c_float, c_int, c_uint, c_uint32, c_int64, c_long, c_ulong, c_ulonglong, c_void_p, c_char_p, CFUNCTYPE, byref, create_string_buffer, addressof, pointer, cast
from scipy.integrate import quad
import math
import numpy as np
from scipy.optimize import leastsq
from scipy.special import poch,factorial2,binom,factorial,gamma,hyp2f1
from collections import defaultdict
import warnings

def get_df_term_latex(k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4,l1,l2,indexIn,indexOut):
    r"""
    Get the latex expression for the disturbing function coefficient

    .. math::
        C_{\pmb{k}}^{\pmb{\nu},\pmb{l}}(\alpha_{i,j})

    Returns
    -------
    latex
    """
    assert l1 + l2 < 2, "I'm not going to print something so ridiculuous!"
    C  = r"\tilde{{C}}_{{({0},{1},{2},{3},{4},{5})}}".format(k1,k2,k3,k4,k5,k6)
    if l1==1:
        C = r"\alpha\frac{d}{d\alpha}" + C
    if l2==1:
        C = r"\alpha\frac{d}{d\alpha}" + C
    C += r"^{{({0},{1},{2},{3})}}".format(nu1,nu2,nu3,nu4)
    C += r"(\alpha_{{{0},{1}}})".format(indexIn,indexOut)
    term = r"{0}\frac{{Gm_{1}m_{2}}}{{a_{{{2},0}}}}".format(["-","+"][l2],indexIn,indexOut)
    term += C
    e1exp = abs(k3) + 2*nu3
    e2exp = abs(k4) + 2*nu4
    s1exp = abs(k5) + 2*nu1
    s2exp = abs(k6) + 2*nu2
    if l1==1:
        term+=r"\left(\frac{{\delta a_{0}}}{{a_{{{0},0}}}}\right)".format(indexIn)
    if l2==1:
        term+=r"\left(\frac{{\delta a_{0}}}{{a_{{{0},0}}}}\right)".format(indexOut)
    if e1exp > 0:
        if e1exp == 1:
            term += r"e_{0}".format(indexIn)
        else:
            term += r"e_{0}^{1}".format(indexIn, e1exp)
    if e2exp > 0:
        if e2exp == 1:
            term += r"e_{0}".format(indexOut)
        else:
            term += r"e_{0}^{1}".format(indexOut, e2exp)
    if s1exp > 0:
        if s1exp == 1:
            term += r"s_{0}".format(indexIn)
        else:
            term += r"s_{0}^{1}".format(indexIn, s1exp)
    if s2exp > 0:
        if s2exp == 1:
            term += r"s_{0}".format(indexOut)
        else:
            term += r"s_{0}^{1}".format(indexOut, s2exp)
    arg = r""
    if k1 == 0 and k2 == 0 and k3 == 0 and k4 == 0 and k5==0 and k6 == 0:
        pass # don't write cosine
    else:
        if k1 < 0:
            arg += r"-"
        if abs(k1) > 1:
            arg += r"{0}".format(abs(k1))
        if k1 != 0:
            arg += r"\lambda_{0}".format(indexOut)
        if k2 != 0:
            arg += r"+" if k2 > 0 else r"-"
            if abs(k2) > 1:
                arg += r"{0}".format(abs(k2))
            arg += r"\lambda_{0}".format(indexIn)
        if k3 != 0:
            arg += r"+" if k3 > 0 else r"-"
            if abs(k3) > 1:
                arg += r"{0}".format(abs(k3))
            arg += r"\varpi_{0}".format(indexIn)  
        if k4 != 0:
            arg += r"+" if k4 > 0 else r"-"
            if abs(k4) > 1:
                arg += r"{0}".format(abs(k4))
            arg += r"\varpi_{0}".format(indexOut)
        if k5 != 0:
            arg += r"+" if k5 > 0 else r"-"
            if abs(k5) > 1:
                arg += r"{0}".format(abs(k5))
            arg += r"\Omega_{0}".format(indexIn)
        if k6 != 0:
            arg += r"+" if k6 > 0 else r"-"
            if abs(k6) > 1:
                arg += r"{0}".format(abs(k6))
            arg += r"\Omega_{0}".format(indexOut)
        term += r"\cos({0})".format(arg)
        
    return term

def get_df_coefficient_symbol(k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4,l1,l2,indexIn,indexOut):
    r"""
    Get a sympy symbol for the disturbing function coefficient

    .. math::
        C_{\pmb{k}}^{\pmb{\nu},\pmb{l}}(\alpha_{i,j})

    Returns
    -------
    sympy symbol
    """
    symbol_str  = r"C_{{({0}\,{1}\,{2}\,{3}\,{4}\,{5})}}".format(k1,k2,k3,k4,k5,k6)
    symbol_str += r"^{{({0}\,{1}\,{2}\,{3})\,({4}\,{5})}}".format(nu1,nu2,nu3,nu4,l1,l2)
    symbol_str += r"(\alpha_{{{0}\,{1}}})".format(indexIn,indexOut)
    return symbols(symbol_str)

def _delta(*args):
    """
    Return 0 if all arguments are 0, otherwise return 1.
    """
    intarr  = np.array([args],dtype=np.int64)
    if np.alltrue(intarr == 0):
        return 0
    else:
        return 1
    
def df_arguments_dictionary(Nmax):
    r"""
    Get the arguments appearing in the disturbing function
    up to order Nmax. Arguments are returned as a nested dictionary.
    The outer level keys are orders. 
    The inner level keys are 
    :math:`\Delta k = |k_1-k_2|`, 
    denoting the order of MMR for which the argument appears. 
    The values of the inner level dictionaries are lists of
    tuples containing :math:`(k_3,k_4,k_5,k_6)`.
    Specifically, the dictionary is returned in the form:
    
    | {
    |     0:{0:[(0,0,0,0)],
    |     1:{1:[(-1,0,0,0),(0,-1,0,0)]},
    |     ...
    |     Nmax:{
    |         0:[(k3,k4,k5,k6)],
    |         ...
    |         Nmax:[(k3,k4,k5,k6),...]
    |     }
    | }
    
    Arguments
    ---------
    Nmax: int
        Maximum order of dictionary arguments.

    Returns
    -------
    arguments : defaultdict
        Nested defaultdicts containing the cosine
        arguments appearing in the disurbing function.
    """
    args_dict = defaultdict(lambda: defaultdict(list))
    Nmax_by_2 = Nmax // 2
    for h in range(Nmax_by_2 + 1):
        khi = 2 * Nmax_by_2
        klo = -khi * _delta(h)
        for k in range(klo,khi + 1):
            hplusk = (h+k)
            hminusk = (h-k)
            s_hi = Nmax - abs(hplusk) - abs(hminusk)
            s_lo = -s_hi * _delta(h,k)
            for s in range(s_lo,s_hi + 1):
                s1_hi = Nmax - abs(hplusk) - abs(hminusk) - abs(s)
                s1_lo = -1 * s1_hi * _delta(h,k,s)
                for s1 in range(s1_lo,s1_hi + 1):
                    dj = 2 * h + s + s1
                    sgn = 1 if dj == 0 else np.sign(dj)
                    k3=-sgn*s
                    k4=-sgn*s1
                    k5=-sgn*(hplusk)
                    k6=-sgn*(hminusk)
                    N = abs(k3) + abs(k4) + abs(k5) + abs(k6)
                    args_dict[N][dj*sgn].append((k3,k4,k5,k6))
    return args_dict

def _nucombos(nutot):
    nucombos = []
    for nu1 in range(nutot+1):
        for nu2 in range(nutot+1-nu1):
            for nu3 in range(nutot+1-nu1-nu2):
                nu4 = nutot - nu1 - nu2 - nu3
                nucombos.append((nu1, nu2, nu3, nu4))
    return nucombos

def _lcombos(ltot):
    lcombos = []
    for l1 in range(ltot+1):
        l2 = ltot - l1
        lcombos.append((l1, l2))
    return lcombos
def k_nu_depend_on_inclinations(k,nu):
    _,_,_,_,k5,k6 = k
    nu1,nu2,_,_ = nu
    arr=np.array([k5,k6,nu1,nu2])
    return np.any(arr!=0)
def k_nu_depend_on_eccentricities(k,nu):
    _,_,k3,k4,_,_ = k
    _,_,nu3,nu4 = nu
    arr=np.array([k3,k4,nu3,nu4])
    return np.any(arr!=0)
def list_resonance_terms(p,q,min_order=None,max_order=None,eccentricities=True,inclinations=True):
    """
    Generate the list of disturbing function terms for a
    p:p-q resonance with eccentricity/inclination order
    between min_order and max_order

    Arguments
    ---------
    p : int
        Determines resonance
    q : int
        Order of the resonance
    min_order : int, optional
        Minimum order in eccentricities and inclinations to include. Defaults to order of the resonance q (leading order)
    max_order: int
        Maximum order in eccentricities and inclinations to include. Defaults to order of the resonance q (leading order)
    eccentricities: bool, optional
        By default, includes all eccentricity terms.
        Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
    inclinations: bool, optional
        By default, includes all inclination terms.
        Can set to False to exclude any inclination terms (e.g., co-planar systems).

    Returns
    -------
    term : list
        A list of disturbing function terms.
        Each entry in the list is of the form
        (k_vec, nu_vec) (see PoincareHamiltonian.add_cosine_term in poincare.py)
    """
    if not min_order:
        min_order = q
    if not max_order:
        max_order = q
    args_dict = df_arguments_dictionary(max_order)
    args = []
    for N in range(min_order,max_order+1):
        for q1 in range(q,N+1,q):
            if (N-q1) % 2:
                continue
            p1 = (q1//q) * p
            for N1 in range(q1,N+1,2):
                nutot = (N-N1)//2
                for arg in args_dict[N1][q1]:
                    for nu_vec in _nucombos(nutot):
                        k_vec = (p1,q1 - p1,*arg)
                        if inclinations == False and k_nu_depend_on_inclinations(k_vec, nu_vec) == True:
                            continue
                        if eccentricities == False and k_nu_depend_on_eccentricities(k_vec, nu_vec) == True:
                            continue
                        args.append((k_vec,nu_vec))
    return args

def list_secular_terms(min_order,max_order,eccentricities=True,inclinations=True):
    """
    Generate the list of secular disturbing function terms 
    with eccentricity/inclination order between min_order and max_order.

    Arguments
    ---------
    min_order : int
     Minimum order in eccentricities and inclinations to include.
    max_order : int
     Maximum order in eccentricities and inclinations to include.
    eccentricities: bool, optional
        By default, includes all eccentricity terms.
        Can set to False to exclude any eccentricity terms (e.g., fixed circular orbits).
    inclinations: bool, optional
        By default, includes all inclination terms.
        Can set to False to exclude any inclination terms (e.g., co-planar systems).

    Returns
    -------
    term : list
        A list of disturbing function terms. 
        Each entry in the list is of the form
        (k_vec, nu_vec) (see PoincareHamiltonian.add_cosine_term in poincare.py)
    """
    args_dict = df_arguments_dictionary(max_order)
    args = []
    Nmax1 = (max_order//2) * 2 
    Nmin1 = (min_order//2) * 2 
    for N in range(0,Nmax1 + 1,2):
        argsN = args_dict[N][0]
        nutot_min = max( (Nmin1 - N)//2 , 0)
        nutot_max = (Nmax1 - N)//2 
        for nutot in range(nutot_min,nutot_max + 1):
            for nu_vec in _nucombos(nutot):
                for arg in argsN:
                    k_vec = (0,0,*arg)
                    if inclinations == False and k_nu_depend_on_inclinations(k_vec, nu_vec) == True:
                        continue
                    if eccentricities == False and k_nu_depend_on_eccentricities(k_vec, nu_vec) == True:
                        continue
                    args.append((k_vec,nu_vec))
    return args

def laplace_b(s,j,n,alpha):
    r"""
    Calculates :math:`n`th derivative with respect to :math:`\alpha` of Laplace coefficient :math:`b_s^j(\alpha)`.
    Uses recursion and scipy special functions. 

    Arguments
    ---------
    s : float 
        half-integer parameter of Laplace coefficient. 
    j : int 
        integer parameter of Laplace coefficient. 
    n : int 
        Specify the :math:`n`th derivative with respect to :math:`alpha`. 
    alpha : float
        Semi-major axis ratio, :math:`\alpha=a_{in}/a_{out}`.

    Returns
    -------
    float
        The value of the coefficient.
    """
    assert alpha>=0 and alpha<1, "alpha not in range [0,1): alpha={}".format(alpha)
    if j<0:
        return laplace_b(s,-j,n,alpha)
    if n >= 2:
        return s * (
            laplace_b(s+1,j-1,n-1,alpha) 
            -  2 * alpha * laplace_b(s+1,j,n-1,alpha)
            + laplace_b(s+1,j+1,n-1,alpha)
            - 2 * (n-1) * laplace_b(s+1,j,n-2,alpha)
        )
    if n==1:
        return s * (
            laplace_b(s+1,j-1,0,alpha) 
            - 2 * alpha * laplace_b(s+1,j,0,alpha) 
            + laplace_b(s+1,j+1,0,alpha)
        )
    return 2 * poch(s,j) * alpha**j * hyp2f1(s,s+j,j+1,alpha**2)/ factorial(j)

def evaluate_df_coefficient_dict(coeff_dict,alpha):
    r"""
    Evaluate a dictionary representing a sum
    of Laplace coefficient terms like those returned
    by df_coefficient_C and df_coefficient_Ctilde evaluated at a 
    specific value of semi-major axis ratio, alpha.

    Arguments
    ---------
    coeff_dict : dictionary
        Dictionary with entries {(p,(s,j,n)) : coeff}
        representing a sum of Laplace coefficients:

        .. math::
         \mathrm{coeff} \times \alpha^p \frac{ d^n b_s^{(j)}(\alpha)} { d\alpha^n}
    alpha : float
        Value of semi-major axis ratio, :math:`\alpha=a_i/a_j`, appearing
        as an argument of Laplace coefficients.

    Returns
    -------
    float :
        The sum of Laplace coefficeint terms represented
        by dictionary entries.
    """
    tot = 0
    for key,val in coeff_dict.items():
        if key[0] == 'indirect':
            pwer = key[1]
            rt_alpha_inv = 1 / np.sqrt(alpha) 
            tot += val * rt_alpha_inv**pwer
        else:
            p,arg = key
            tot += val * alpha**p * laplace_b(*arg,alpha)
    return tot


def calB(n,k,p):
    arglist = [negative_binom(p,l) * factorial(l) for l in range(1,n-k+3)]
    return bell(n,k,arglist)

def falling_factorial(x,n):
    return poch(-x,n) * (-1)**n

def _Psi_coeff(l1,l2,p1,p2,m1,m2,r1,r2):
    return binom(l1,m1) * binom(l2,m2) *\
    falling_factorial(-2*r1 - p2/2,m2) * falling_factorial(-p1/2,m1) *\
    calB(l1-m1,r1,2) * calB(l2-m2,r2,-2)

def _calc_df_coefficient_C_l1_l2_Taylor_coeff(l1,l2,p1,p2,derivs_list):
    tot = 0
    l1fact_inv = 1/factorial(l1)
    l2fact_inv = 1/factorial(l2)
    prefactor = l1fact_inv * l2fact_inv
    for m1 in range(l1+1):
        for r1 in range(l1-m1+1):
            for m2 in range(l2+1):
                for r2 in range(l2-m2+1):
                    val = prefactor * derivs_list[r1+r2] * _Psi_coeff(l1,l2,p1,p2,m1,m2,r1,r2)
                    tot=tot+val        
    return tot

def _p1_p2_from_k_nu(kvec,nuvec):
    r"""
    Convenience function for calculate integers p1 and p2 
    defined as 
        p1 = abs(k3) + abs(k5) + 2 * nu1 + 2 * nu3
        p2 = 4 + abs(k4) + abs(k6) + 2 * nu2 + 2 * nu4

    These definitions appear in the formula for the expansion
    of DF coefficients in powers delta
    """
    k1,k2,k3,k4,k5,k6 = kvec
    nu1,nu2,nu3,nu4 = nuvec
    p1 = abs(k3) + abs(k5) + 2 * nu1 + 2 * nu3
    p2 = 4 + abs(k4) + abs(k6) + 2 * nu2 + 2 * nu4
    return p1,p2

def evaluate_df_coefficient_delta_expansion(Coeff_dict,p1,p2,lmax,alpha):
    r"""
    Calculate coefficients of the Taylor series of `Coeff_dict` in 
    powers math:`\delta_i` and math::`\delta_j` evaluated at semi-major
    axis ratio math:`\alpha=a_1/a_2`.

    Arguments
    ---------
    Coeff_dict : dict
        Dictionary representation of DF coefficient
    p1 : int
        Integer specific to the DF term's (k,nu) values
        given by
            p1 = abs(k3) + abs(k5) + 2 * nu1 + 2 * nu3
    p2 : int
        Integer specific to the DF term's (k,nu) values
        given by
            p2 = 4 + abs(k4) + abs(k6) + 2 * nu2 + 2 * nu4
    lmax : int
        Maximum power of Taylor expansion.
    alpha : float
        Semi-major axis ratio at which to evaluate coefficients.
    """
    answer = dict()
    C = Coeff_dict
    # Derivatives of C w.r.t. alpha
    NC_derivs = [evaluate_df_coefficient_dict(C,alpha)]
    for ltot in range(lmax+1):
        C=deriv_df_coefficient(C)
        NC_derivs.append(alpha**(ltot+1) * evaluate_df_coefficient_dict(C,alpha))
        for lIn in range(ltot+1):
            lOut = ltot - lIn
            answer[(lIn,lOut)] = _calc_df_coefficient_C_l1_l2_Taylor_coeff(lIn,lOut,p1,p2,NC_derivs)
    return answer

# Vector of resonance coefficients
def get_res_coefficient_vector(j,k,include_indirect_terms=True):
    r"""
    Get a vector comprised of all sub-resonance coefficients for the j:j-k mean motion resonance.

    Arguments
    ---------
    j : int
        Specify the j:j-k resonance
    k : int
        Order of the resonance
    include_indirect_terms :  boole, optional
        whether the contribution of indirect terms should be
        accounted for when computing the coefficients. default
        is true.

    Returns
    -------
    vals : ndarray
        Array of coefficient values from l=0 to l=k
    """ 
    res_pratio = float(j - k) /float(j)
    alpha = res_pratio**(2./3.)
    res_ecc_arg = lambda j,k,l: (j,k-j,-l,l-k,0,0,0,0,0,0) # k and nu values
    Cjkl = lambda j,k,l,alpha: evaluate_df_coefficient_dict(df_coefficient_C(*res_ecc_arg(j,k,l),include_indirect_terms=include_indirect_terms),alpha)
    vals = np.array([Cjkl(j,k,l,alpha) for l in range(k+1)],dtype=np.float64)
    return vals

def get_fg_coefficients(res_j,res_k):
    """Get 'f' and 'g' coefficients for approximating the disturbing function coefficients associated with an MMR."""
    res_pratio = float(res_j - res_k) /float(res_j)
    alpha = res_pratio**(2./3.)
    vec = get_res_coefficient_vector(res_j,res_k)
    resids_vec_fn = lambda fg: vec - np.array([binomial(res_k,l) * fg[0]**(l) * fg[1]**(res_k-l) for l in range(res_k+1)],dtype=np.float64)
    ex = (1-alpha)
    f0 = -1 / ex
    g0 = 1 / ex
    f,g = leastsq(resids_vec_fn,(f0,g0))[0]
    return f,g


def NCOd0(a,b,c):
    """
    Value of of Newcomb operator
        X^{a,b}_{c,0}
        
    Arguments
    ---------
    a : int
    b : int
    c : int
    
    Returns
    -------
    X^{a,b}_{c,0} : float
    """
    if c==0: return 1
    if c==1: return b - a /2
    nc1 = NCOd0(a,b+1,c-1)
    nc2 = NCOd0(a,b+2,c-2)
    return  (2 * (2*b-a) * nc1 + (b-a) * nc2 )/ c / 4

def NewcombOperator(a,b,c,d):
    """
    Value of of Newcomb operator
        X^{a,b}_{c,d}
        
    Arguments
    ---------
    a : int
    b : int
    c : int
    d : int
    
    Returns
    -------
    X^{a,b}_{c,d} : float
    """
    if c<0 or d<0: return 0
    if d==0: return NCOd0(a,b,c)
    tot = -2 * (2 * b + a) * NewcombOperator(a,b-1,c,d-1)
    tot += -1 * (b + a) * NewcombOperator(a,b-2,c,d-2)
    tot += -1 * (c - 5 * d + 4 + 4 * b + a ) * NewcombOperator(a,b,c-1,d-1)    
    for j in range(2,d+1):
        tot += 2 * (c-d+b) * (-1)**j * binom(3/2,j) * NewcombOperator(a,b,c-j,d-j)
    return tot / 4 / d

def HansenCoefficient_term(a,b,c,sigma):
    r"""
    Series coefficient in Taylor series
    of the Hansen coefficient X^{a,b}_c(e).
    The Hansen coefficient is given by:

    .. math::

        X^{a,b}_c(e) = e^{|c-b|} \times
        \sum_{\sigma=0}^\infty \mathrm{HansenCoefficient\_term(a,b,c,sigma)}e^{2\sigma}

    Arguments
    ---------
    a : int

    b : int
    
    c : int
    
    sigma : int

    Returns
    -------
    float
    """
    alpha = max(c-b,0)
    beta  = max(b-c,0)
    return NewcombOperator(a,b,alpha+sigma,beta+sigma)

def calX_term(a,b,c,d):
    r"""
    Series coefficient in Taylor series of the expession 
    calX^{a,b}_c(e) = (1-e^2)^{-1/2} * X^{a,b}_c(e) where
    X^{a,b}_c is a traditional Hansen coefficient 
    math::
        calX^{a,b}_c(e) = e^{|c-b|} \times
        \sum_{\sigma=0}^\infty calX_term(a,b,c,sigma)e^{2\sigma}

    Arguments
    ---------
    a : int
    
    b : int
    
    c : int
    
    d : int

    Returns
    -------
    float
    """
    tot = 0
    for n in xrange(d+1):
        tot += binom(-0.5,n) * (-1)**n * HansenCoefficient_term(a,b,c,d - n)
    return tot

def threeFtwo(a,b):
    """
    Hypergerometric 3_F_2([a1,a2,a3],[b1,b2],1)
    
    Used in calcluations of KaulaF function
    
    Arguments
    ---------
    a : list of ints
    
    b :  list of ints
    
    Returns
    -------
    float
    """
    a1,a2,a3 = a
    b1,b2 = b
    kmax = min(1-a1,1-a2,1-a3)
    tot = 0
    for k in range(0,kmax):
        tot += poch(a1,k) * poch(a2,k) *poch(a3,k) / poch(b1,k) / poch(b2,k) / factorial(k)
    return tot

def KaulaF(n,q,p,j):
    """
    Series coefficient in the Taylor expansion of the Kaula inclination
    function F_{nqp}(I).  See, e.g., `Kaula (1962,1966)`_ or `Ellis & Murray
    (2000)`_.

    The function returns the jth term of the Taylor
    expansion in the variable s = sin(I/2). I.e.
        KaulaF(n,q,p,j) = (1/j!) d^j F{nqp}/ ds^j

    This implementation is based on the Mathematica
    package by Fabio Zugno available at:
        https://library.wolfram.com/infocenter/MathSource/4256/

    .. _Kaula (1962,1966): https://ui.adsabs.harvard.edu/abs/1962AJ.....67..300K/abstract
    .. _Ellis & Murray (2000): https://ui.adsabs.harvard.edu/abs/2000Icar..147..129E/abstract
    Arguments
    ---------
    n : int

    q : int
    
    p : int
    
    j : int

    Returns
    -------
    float
    """
    if q==0 and 2 * p == n:
        return (-1)**(j+n) * binom(n,j) * binom(n+j,j) * factorial2(n-1) / factorial2(n)
    if n - 2*p - q < 0: 
        if n-q<0: return 0
        return (-1)**(n-q) * factorial(n+q) * KaulaF(n,-q,n-p,j) / factorial(n-q)
    numerator =  (-1)**j * factorial2(2*n-2*p-1)     
    numerator *= binom(n/2+p+q/2,j) 
    numerator *= threeFtwo([-j,-2*p,-n-q],[1+n-2*p-q,-(n/2)-p-q/2]) 
    if p<0: return 0.
    denom =  factorial(n -2 * p - q) * factorial2(2 * p)
    return numerator / denom 

def KK(i,n,m):
    numerator = (-1)**(i-n) * (1 + 2 * n) * gamma(1/2 + i) * gamma(3/2 + i)
    denom = 4 * gamma((2 + i-m-n)/2) * gamma((2 + i + m-n)/2) * gamma((3 + i - m + n)/2) * gamma((3 + i + m + n)/2)
    return numerator / denom

def getrange(lim1,lim2,n=1):
    assert n>0, "Negative interval n={} passed to getrange".format(n)
    if lim1 < lim2:
        return range(lim1,lim2 + n,n)
    else:
        return range(lim1,lim2-n,-n)
    
def FX(h,k,i,p,u,v1,v2,v3,v4,z1,z2,z3,z4):
    
    nlim1 = int(np.ceil(max(h,(h + abs(k)) / 2)))
    nlim2 = i
    hplusk_mod2 = (h+k) % 2
    delta = 1 - np.alltrue(np.array([h,k,v1-v2,v4-v3]) == 0)
    inc_total= 0
    for n in getrange(nlim1,nlim2,1):
        mlim1 = max(n-i,p-n+h,p-k-n+hplusk_mod2)
        mlim2 = min(i-n,p+n-h,n+p-k-hplusk_mod2)
        for m in getrange(mlim1,mlim2,2):
            term = (-1)**abs(k+m+n-p) * KK(i,n,m)
            term *= KaulaF(n,-k-m+p,(-h + m + n - p)//2,z1)
            term *= KaulaF(n, k+m-p,(-h - m + n + p)//2,z2)
            inc_total += term
    
    ecc_total = 0
    for t in range(0,u+1):
        term = (-1)**(u + t) / factorial(u-t) / factorial(t)
        term *= HansenCoefficient_term(i+t,v1,v2,z3)
        term *= HansenCoefficient_term(-1-i-t,v3,v4,z4)
        ecc_total += term
        
    return (1 + delta) * inc_total * ecc_total

    

def df_coefficient_Ctilde(k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4,include_indirect = True):
    r"""
    Get the coefficient of the disturbing function term:

    .. math::
        s_i^{|k_5|+2\nu_1} s_j^{|k_6|+2\nu_2}
        e_i^{|k_3|+2\nu_4} e_j^{|k_4|+2\nu_3} 
        \times
        \cos[
            k_1\lambda_j + k_2\lambda_i +
            k_3 \varpi_i + k_4 \varpi_j +
            k_5 \Omega_i + k_6 \Omega_j
        ]

    where the indices :math:`i` and :math:`j` corresponds to the inner and
    outer planets, respectively.  The result is returned as a dictionary of
    Laplace coefficient arguemnts and their numerical coefficents.

    Arguments:
    ----------
    k1 : int
        Coefficient of outer planet's mean longitude in cosine argument
    k2 : int
        Coefficient of inner planet's mean longitude in cosine argument
    k3 : int
        Coefficient of inner planet's mean longitude in cosine argument
    k4 : int
        Coefficient of outer planet's mean longitude in cosine argument
    k5 : int
        Coefficient of inner planet's longitude of ascending node in cosine argument
    k6 : int
        Coefficient of outer planet's longitude of ascending node in cosine argument
    nu1 : int
        Select specific term where the exponent of :math:`s_i` is :math:`|k_5|+2\nu_1`
    nu2 : int
        Select specific term where the exponent of :math:`s_j` is :math:`|k_6|+2\nu_2`
    nu3 : int
        Select specific term where the exponent of :math:`e_i` is :math:`|k_3|+2\nu_3`
    nu4 : int
        Select specific term where the exponent of :math:`e_j` is :math:`|k_4|+2\nu_4`
    include_indirect : booole, optional
        Whether to include the indirect contribution to the disturibing function
        coefficient.

    Returns
    -------
    dictionary
        The coefficient is given by the sum over laplace coefficients
        contained in the dictionary entries: 

        ..math::
            \sum A \times \alpha^p \frac{d^{n}}{d\alpha^{n}} b_{s}^{j}(\alpha)

        where the dictionary entries are in the form { (p,(s,j,n)) : A }
    """
    total = defaultdict(float)
    # must be even power in inclination
    if abs(k5 + k6) % 2:
        warnings.warn(
                "\n df_coefficient called with an argument not symmetric w.r.t. planet inclinations:\n" + 
                "\t (k1,k2,k3,k4,k5,k6)=({},{},{},{},{},{})".format(k1,k2,k3,k4,k5,k6)
                )
        return dict(total)
    # Sum of integer coefficients must be 0
    if k1 + k2 + k3 + k4 + k5 + k6:
        warnings.warn(
                "\n df_coefficient called with an argument that does not satisfy D'Alembert relation:\n" + 
                "\t (k1,k2,k3,k4,k5,k6)=({},{},{},{},{},{})".format(k1,k2,k3,k4,k5,k6)
                )
        return dict(total)
    kvec = np.array([k1,k2,k3,k4,k5,k6])
    if np.alltrue(kvec==0):
        for i in getrange(0,nu1+nu2,1):
            for p in getrange(i%2,i,2):
                for u in getrange(0,2*nu3+2*nu4):
                    cf = FX(0,0,i,p,u,0,0,0,0,nu1,nu2,nu3,nu4) * (1 + (p != 0))
                    if not np.isclose(cf,0):
                        total[(i+u,(i+1/2,abs(p),u))]+=cf
    elif np.alltrue(kvec[2:]==0):
        j = abs(k1)
        for i in getrange(0,nu1+nu2,1):
            for p in getrange(i%2,i,2):
                for u in getrange(0,2*nu3+2*nu4):
                    cf = FX(0,0,i,p,u,j,j,j,j,nu1,nu2,nu3,nu4) 
                    if not np.isclose(cf,0):
                        total[(i+u,(i+1/2,abs(j+p),u))]+=cf
                        if p != 0:
                            total[(i+u,(i+1/2,abs(j-p),u))]+=cf
    else:
        h = (-k5-k6)//2
        k = (k6-k5)//2
        n0 = int(np.ceil(max(h,-k5/2,-k6/2))) 
        for i in getrange(n0,nu1 + nu2 + (abs(k5)+abs(k6)//2),1):
            for p in getrange(h-i,i-h,2):
                for u in getrange(0,2*nu3 + 2*nu4 + abs(k3) + abs(k4),1):
                    cf = FX(h, k, i, p, u, k2 + k3, k2, k1 + k4, k1, nu1, nu2, nu3, nu4)
                    if not np.isclose(cf,0):
                        total[(i+u,(i+1/2,abs(k1+k4-h+p),u))]+=cf

    # add indirect term
    if include_indirect:
        coeff = df_coefficient_Ctilde_indirect_piece(k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4)
        total[('indirect',1)] = coeff
    return dict(total)

def _df_coeff_mult_by_alpha_power(coeffdict,r):
    """
    For a DF coefficient 'coeff' represented by 'coeffdict', 
    return the new coefficient dictionary  alpha^r * coeff
    """
    new = dict()
    for key,val in coeffdict.items():
        if key[0]=='indirect':
            newkey = (key[0],key[1]-2*r)
        else:
            p,sjn = key
            newkey = (p+r,sjn)
        new[newkey]=val
    return new

def _df_coefficient_scalar_mult(coeffdict,s):
    """
    Return dictionary for s * coeffdict for scalar value 's'
    """
    return {key:s*val for key,val in coeffdict.items()}

def _get_alpha_times_derivs_list(coeffdict, ltot):
    derivs_list=[coeffdict]
    dcoeff = coeffdict.copy()
    for l in range(ltot):
        dcoeff = deriv_df_coefficient(dcoeff)
        derivs_list.append(_df_coeff_mult_by_alpha_power(dcoeff,l+1))
    return derivs_list



def df_coefficient_C(k1,k2,k3,k4,k5,k6,nu1,nu2,nu3,nu4,l1=0,l2=0,include_indirect_terms = True):
    r"""
    Get the coefficient of the disturbing function term

    .. math ::

        |Y_i|^{|k_5|+2\nu_1}
        |Y_j|^{|k_6|+2\nu_2}
        |X_i|^{|k_3|+2\nu_3}
        |X_j|^{|k_4|+2\nu_4}
        \delta_i^{l_1}
        \delta_j^{l_2}
        \times \cos[
            k_1 \lambda_j + k_2 \lambda_i +
            k_3 \varpi_i + k_4 \varpi_j +
            k_5 \Omega_i + k_6 \Omega_j
        ]~,

    where the indices :math:`i` and :math:`j` corresponds to the
    inner and outer planets, respectively.
    The result it returned as a dictionary of Laplace coefficient
    arguemnts and their numerical coefficents.

    Arguments:
    ----------
    k1 : int
        Coefficient of outer planet's mean longitude in cosine argument
    k2 : int
        Coefficient of inner planet's mean longitude in cosine argument
    k3 : int
        Coefficient of inner planet's mean longitude in cosine argument
    k4 : int
        Coefficient of outer planet's mean longitude in cosine argument
    k5 : int
        Coefficient of inner planet's longitude of ascending node in cosine argument
    k6 : int
        Coefficient of outer planet's longitude of ascending node in cosine argument
    nu1 : int
        Select specific term where the exponent of :math:`|Y_i|` is :math:`|k_5|+2\nu_1`
    nu2 : int
        Select specific term where the exponent of :math:`|Y_j|` is :math:`|k_6|+2\nu_2`
    nu3 : int
        Select specific term where the exponent of :math:`|X_i|` is :math:`|k_3|+2\nu_3`
    nu4 : int
        Select specific term where the exponent of :math:`|X_j|` is :math:`|k_4|+2\nu_4`
    l1 : int, optional
        Select specifc term where exponent of 
        :math:`\delta_1 = (\Lambda_1 - \Lambda_{1,0})/\Lambda_{1,0})`
        is l1.
        Default value is l1 = 0
    l2 : int, optional
        Select specifc term where exponent of 
        :math:`\delta_2 = (\Lambda_2 - \Lambda_{2,0})/\Lambda_{2,0})`
        is l2.
        Default value is l2 = 0
    include_indirect_terms :  boole, optional
        whether the contribution of indirect terms should be
        accounted for when computing the coefficients. default
        is true.

    Returns
    -------
    dictionary
        The coefficient is given by the sum over laplace coefficients
        contained in the dictionary entries:

        .. math::
            \sum C \times \alpha^p \frac{d^{n}}{d\alpha^{n}} b_{s}^{j}(\alpha)

        where the dictionary entries are in the form { (p,(s,j,n)) : C }
    """
    Chat = defaultdict(float)
    msg = "Integer arguments nu_i and l_i must be non-negative."
    assert np.alltrue(np.array([nu1,nu2,nu3,nu4,l1,l2])>=0), msg
    for n3 in range(nu3+1):
        for n4 in range(nu4+1):
            term_dict = df_coefficient_Ctilde(k1,k2,k3,k4,k5,k6,nu1,nu2,n3,n4,include_indirect_terms)
            prefactor = Xi(nu3-n3,n3+abs(k3)/2,nu1+abs(k5)/2) * Xi(nu4-n4,n4+abs(k4)/2,nu2+abs(k6)/2)
            if prefactor != 0.:
                for key,val in term_dict.items():
                    Chat[key] += prefactor * val
    ltot = l1 + l2
    if ltot==0:
        result = Chat
    else:
        result = defaultdict(float)
        derivs_list = _get_alpha_times_derivs_list(Chat,ltot)
        l1fact_inv = 1/factorial(l1)
        l2fact_inv = 1/factorial(l2)
        prefactor = l1fact_inv * l2fact_inv
        p1,p2= _p1_p2_from_k_nu([k1,k2,k3,k4,k5,k6],[nu1,nu2,nu3,nu4])
        for m1 in range(l1+1):
            for r1 in range(l1-m1+1):
                for m2 in range(l2+1):
                    for r2 in range(l2-m2+1):
                        to_add = _df_coefficient_scalar_mult(derivs_list[r1+r2],prefactor * _Psi_coeff(l1,l2,p1,p2,m1,m2,r1,r2))
                        result = _add_dicts(result,to_add)      
    return result

def has_indirect_component(k1,k2,k3,k4,k5,k6):
    two_p = k2 + k4 + 1 
    two_p1 = -1 * (k1 + k3 - 1)
    m = k5 - two_p1 + 1
    m_is_zero_or_one = m == 0 or m == 1
    return is_zero_or_two(two_p) and is_zero_or_two(two_p1) and m_is_zero_or_one

def is_zero_or_two(n):
    return n == 0 or n == 2
    
def deriv_df_coefficient(coeff):
    r"""
    Derivative of a disturbing function coefficient with repsect to 
    alpha.

    Argument
    --------
    coeff : dict
        The coefficient is given by the sum over laplace coefficients
        contained in the dictionary entries:

            .. math::
            \sum C \times \alpha^p \frac{d^{n}}{d\alpha^{n}} b_{s}^{j}(\alpha)
        where the dictionary entries are in the form { (p,(s,j,n)) : C }
    
    Returns
    -------
    dict 
        A new dictionary in the same form as the input 'coeff' representing
        the derivative of coeff with respect to alpha.
    """
    dcoeff = defaultdict(float)
    for key,val in coeff.items():
        if key[0]=='indirect':
            # term \propto \alpha^{-p/2}
            pwer = key[1]
            dcoeff[('indirect',pwer + 2)] = -0.5 * pwer * val
        else:
            p,sjn= key
            s,j,n = sjn
            if p>0:
                dcoeff[(p-1,sjn)] += p * val
            dcoeff[(p,(s,j,n+1))] += val
    return dict(dcoeff)

def Xi(N,p,q):
    tot = 0
    for l in range(0,N+1):
        tot += binom(p,l) * negative_binom(-q,N-l) * (1/2)**l
    tot *= (-1/2)**N
    return tot

def negative_binom(minus_q,l):
    # scipy.special.binom returns a NaN when called at a
    # negative integer so I use this alternate formulation
    # when the argument is potenially a negative integer
    return (-1)**l * poch(-1 * minus_q,l) / factorial(l)


def _Cindirect_type1(k1,k2,m,z1,z2,z3,z4):
    r"""
    Indirect term cosine coefficient for argument:
         \theta = k_1\lambda_j + k_2\lambda_i
                    - (k_1-1)\varpi_j - (k_2+1) \varpi_i
                    + m(\Omega_i-\Omega_j)
    """
    if m > 2 or m < 0:
        return 0
    base_case = -1 * calX_term(0,1,-k1,z4) * calX_term(0,1,k2,z3)
    if m == 0:
        return binom(1,z1) * binom(1,z2) * (-1)**(z1) * (-1)**(z2) * base_case
    if m == 1:
         return 2 * binom(0.5 ,z1) * binom(0.5, z2 ) * (-1)**(z1) * (-1)**(z2) *\
        base_case
    if m == 2:
        return (1 - _delta(z1)) * (1 - _delta(z2)) * base_case

def _Cindirect_type2(k1,k2,m,z1,z2,z3,z4):
    r"""
    Indirect term cosine coefficient for argument:

        .. math::
        \theta = k_1\lambda_j + k_2\lambda_i
                    - (k_1-1)\varpi_j - (k_2-1) \varpi_i
                    - \Omega_i - \Omega_j + m(\Omega_j-\Omega_i)
    """
    if m > 2 or m <0:
        return 0
    base_case = -1 * calX_term(0,1,k1,z4) * calX_term(0,1,k2,z3)
    if m == 0:
        return binom(1 ,z1) * (1 - _delta(z2)) * (-1)**(z1) * base_case
    if m == 1:
        return -2 * binom(0.5 ,z1) * binom(0.5, z2 ) * (-1)**(z1) * (-1)**(z2) * base_case
    if m == 2:
        return binom(1 ,z2) * (1 - _delta(z1)) * (-1)**(z2) * base_case


def df_coefficient_Ctilde_indirect_piece(k1,k2,k3,k4,k5,k6,z1,z2,z3,z4):
    r"""
    Get the indirect contribution to disturbing function coefficient

    .. math::
        \bar{C}_{\pmb{k}}^{(z_1,z_2,z_3,z_4)}
    """
    if np.sum([k1,k2,k3,k4,k5,k6]) !=0:
        warnings.warn(
        "\n df_coefficient called with an argument that does not satisfy D'Alembert relation:\n" +
        "\t (k1,k2,k3,k4,k5,k6)=({},{},{},{},{},{})".format(k1,k2,k3,k4,k5,k6)
        )
        return 0
    if k1 == 0 or k2 == 0:
        return 0
    s1 = k1 + k4
    s2 = k2 + k3
    if abs(s1) == 1 and abs(s2)==1:
        if s1 == -1*s2:
            return _Cindirect_type1(s2 * k1,s2 * k2,s2 * k6,z1,z2,z3,z4)
        else:
            return _Cindirect_type2(s1 * k1,s1 * k2,-1 * s1 * k5,z1,z2,z3,z4)
    else:
        return 0

# SH --- Needs modified to accomodate expansion in delta.
def terms_list_to_HamiltonianCoefficients_dict(terms_list,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out,include_alpha_derivs=False):
    """
    Retrieve the a dictionary with the coefficient values of terms appearing in the Hamiltonian.
    
    Arguments
    ---------
    terms_list : list
      List of terms to copmute coefficients for. List entries
      should be in the form (kvec, zvec) where kvec and zvec
      are tuples of integers.
    G : float
      Value of gravitational constant.
    mIn : float
      Inner planet mass parameter
    mOut : float
      Outer planet mass parameter
    MIn : float
      Inner planet stellar mass parameter
    MOut : float
      Outer planet stellar mass parameter
    Lambda0In : float
      Lambda of inner planet for evaluating coefficient.
    Lambda0Out : float
      Lambda of outer planet for evaluating coefficient.
    include_alpha_derivs : bool, optional
      If True, dictionary values are tuple containing
      both the coefficient values and their derivatives
      with respect to alpha = aIn/aOut.
    
    Returns
    -------
    coeff_dictionary : dict
      Coefficients given in dictionary where entries are given in the form
        {(kvec,zvec):Coeff}
      or, if include_alpha_derivs=True, in the form
        {(kvec,zvec):(Coeff,dCoeff/dalpha}
      where 'Coeff' represents the coefficient the term
        (1/2) * Coeff * \exp[i * (k[0] *\lambda_{out} + k[1] *\lambda_{in})] * 
          X_{in}^{|kvec[2]| + zvec[2]} * 
          X_{out}^{|kvec[3]| + zvec[3]} * 
          Y_{in}^{|kvec[4]| + zvec[0]} * 
          Y_{out}^{|kvec[5]| + zvec[1]}
          +
          complex conjugate
      appearing in the interaction Hamiltonian between a pair of planets.
      
    """
    muIn = mIn * (MIn - mIn) / MIn
    muOut = mOut * (MOut - mOut) / MOut
    aIn0 = (Lambda0In / muIn)**2 / MIn / G
    aOut0 = (Lambda0Out / muOut)**2 / MOut / G
    alpha = aIn0 / aOut0
    assert alpha < 1, "Particles are not in order by semi-major axis."
    aOut_inv = G*MOut*muOut*muOut / Lambda0Out / Lambda0Out  
    prefactor = -G * mIn * mOut * aOut_inv
    result = dict()
    for kvec,zvec in terms_list:
        C = df_coefficient_C(*kvec,*zvec)
        coeff = prefactor * evaluate_df_coefficient_dict(C,alpha)
        if include_alpha_derivs:
            ind = C.pop(('indirect',1))
            dC = deriv_df_coefficient(C)
            dcoeff = prefactor * ( evaluate_df_coefficient_dict(dC,alpha) - 0.5 * ind / np.sqrt(alpha*alpha*alpha))
            result[(kvec,zvec)] = coeff,dcoeff
        else:
            result[(kvec,zvec)] = coeff
    return result

def kz_to_xx1yy1_powers(kz):
    """
    Convert (kvec,zvec) Poisson series term designation 
    to the integer powers of variables x_{in},x_{out},y{in}, and y_{out}
    and their complex conjugates.

    Arguments
    ---------
    kz : tuple
      tuple (kvec,zvec) designating the Poisson series term.

    Returns
    -------
    pows : array
      Poisson series powers
      x_{in}^{pows[0]} * 
      x_{out}^{pows[1]} *
      y_{in}^{pows[2]} * 
      y_{out}^{pows[3]} 
      xbar_{in}^{pows[4]} * 
      xbar_{out}^{pows[5]} *
      ybar_{in}^{pows[6]} * 
      ybar_{out}^{pows[7]} 
    """
    k,z = kz
    zpow = lambda x: max(x,0)
    zbarpow = lambda x: max(-x,0)
    xx1yy1_pows = [
        z[2] + zpow(k[2]),
        z[3] + zpow(k[3]),
        z[0] + zpow(k[4]),
        z[1] + zpow(k[5])
    ]
    xx1yy1_bar_pows = [
        z[2] + zbarpow(k[2]),
        z[3] + zbarpow(k[3]),
        z[0] + zbarpow(k[4]),
        z[1] + zbarpow(k[5])
    ]
    return np.array(xx1yy1_pows + xx1yy1_bar_pows,dtype=np.int64)

def xx1yy1_powers_to_kz(xx1yy1_powers,k1=0,k2=0):
    """
    Inverse of 'kz_to_xx1yy1_powers'
    """
    pows = xx1yy1_powers[:4]
    bar_pows = xx1yy1_powers[4:]
    zpows = np.min(np.vstack((pows,bar_pows)),axis=0)
    kvec=np.concatenate(([k1,k2],pows - bar_pows))
    zvec = np.concatenate((zpows[2:],zpows[:2]))
    return tuple(kvec),tuple(zvec)

def xx1yy1_powers_conj(xx1yy1_powers):
    return np.concatenate([xx1yy1_powers[-4:],xx1yy1_powers[:4]])

def _add_ppbar_bracket_terms(coeff1,k1z1,coeff2,k2z2,results,LambdaIn,LambdaOut):
    pows1 = kz_to_xx1yy1_powers(k1z1)
    pows2 = kz_to_xx1yy1_powers(k2z2)
    pows1conj = xx1yy1_powers_conj(pows1)
    pows2conj = xx1yy1_powers_conj(pows2)
    LmbdaInv_factors = np.array((2/LambdaIn,2/LambdaOut,0.5/LambdaIn,0.5/LambdaOut))
    p1conjp2 = pows1conj + pows2
    for i,p1bar,p2 in zip(range(4),pows1conj[:4],pows2[4:]):
        if p1bar > 0 and p2 > 0:
            v = p1conjp2.copy()
            v[i]-=1
            v[4+i]-=1
            coeff = p1bar * p2 * coeff1 * coeff2 * LmbdaInv_factors[i]
            kznew = xx1yy1_powers_to_kz(v)
            results[kznew] -= coeff
    p1p2conj = pows1 + pows2conj
    for i,p1,p2bar in zip(range(4),pows1[:4],pows2conj[4:]):
        if p1 > 0 and p2bar > 0:
            v = p1p2conj.copy()
            v[i]-=1
            v[4+i]-=1
            coeff = p1 * p2bar * coeff1 * coeff2 * LmbdaInv_factors[i]
            kznew = xx1yy1_powers_to_kz(v)
            results[kznew] += coeff

def _consolidate_dictionary_terms(d):
    """
    Combine all terms (kvec,zvec) that are conjugates of one another
    into single terms with a positive coefficient for 
    pomega_in (i.e., k[2]) or, if k[2]=0, Omega_in (k[4]), or, if k[4]=0,
    pomega_out k[3]
    """
    dnew = defaultdict(int)
    for kz, val in d.items():
        k,z = kz
        k = np.array(k,dtype=np.int64)
        if k[2] != 0:
           k *= np.sign(k[2])
        elif k[4] !=0:
           k *= np.sign(k[4]) 
        elif k[3] !=0:
           k *= np.sign(k[3]) 
        else: 
            assert np.alltrue(k==0) and not np.alltrue(np.array(z)==0), "Suspected bad k,z={},{} pair in term dictionary!".format(k,z)
        dnew[(tuple(k),z)] += val
    return dict(dnew)

def resonant_terms_list_to_secular_contribution_dictionary(terms_list,j,k,Nmin,Nmax,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
    """
    Generate a dictionary containing the secular terms generated by a list of resonant terms.
    The secular contributions arise at second order in planet masses.

    Arguments
    ---------
    terms_list : list
      List of resonance terms in (kvec,zvec) form.
    j : int
      Specifies resonance as the j:j-k MMR.
    k : int
      Order of MMR.
    G : float
      Value of gravitational constant.
    mIn : float
      Inner planet mass parameter
    mOut : float
      Outer planet mass parameter
    MIn : float
      Inner planet stellar mass parameter
    MOut : float
      Outer planet stellar mass parameter
    Lambda0In : float
      Lambda of inner planet for evaluating coefficient.
    Lambda0Out : float
      Lambda of outer planet for evaluating coefficient.
    Nmin : int
      Minimum power (in inclination/eccentricity) of terms to include in result.
    Nmax : int
      Maximum power (in inclination/eccentricity) of terms to include in result.

    Returns
    -------
    result : dict
      A dictionary containing the secular terms with 
      entries in the form
       {(kvec,zvec):Coeff}
    """
    muIn = mIn * (MIn - mIn) / MIn
    muOut = mOut * (MOut - mOut) / MOut
    omega_vec = G * G * np.array([
        MIn * MIn * muIn * muIn * muIn / Lambda0In / Lambda0In / Lambda0In,
        MOut * MOut * muOut * muOut * muOut / Lambda0Out / Lambda0Out / Lambda0Out
    ])
    Domega = np.diag(-3 * omega_vec / np.array([Lambda0In,Lambda0Out]))
    res_kvec = np.array([k-j,j])
    res_omega = res_kvec @ omega_vec
    res_factor = 0.25 * (res_kvec @ Domega @ res_kvec) / res_omega / res_omega
    p = terms_list_to_HamiltonianCoefficients_dict(terms_list,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out,include_alpha_derivs=True)
    aIn0 = (Lambda0In / muIn)**2 / MIn / G
    aOut0 = (Lambda0Out / muOut)**2 / MOut / G
    alpha = aIn0 / aOut0
    dalpha_dLambdaIn = 2 * alpha / Lambda0In
    dalpha_dLambdaOut = -2 * alpha / Lambda0Out
    result = defaultdict(int)
    for kz1,coeffs1 in p.items():
        coeff1,dcoeff1 = coeffs1
        for kz2, coeffs2 in p.items():
            coeff2,dcoeff2 = coeffs2
            pows1 = kz_to_xx1yy1_powers(kz1)
            pows2 = kz_to_xx1yy1_powers(kz2)
            pows2conj = xx1yy1_powers_conj(pows2)
            powsnew = pows1 + pows2conj
            tot_pow = np.sum(powsnew)
            if Nmin<=tot_pow<=Nmax:
                rtLambda_inv_pow_In = powsnew[0] + powsnew[2]
                rtLambda_inv_pow_Out = powsnew[1] + powsnew[3]
                newk,newz = xx1yy1_powers_to_kz(powsnew)
                # term 1 part
                val = res_factor*coeff1*coeff2
                # term 2 part
                d_ppbar_dLambdaIn =  (dcoeff1 * coeff2 + coeff1 * dcoeff2) * dalpha_dLambdaIn  - 0.5 * rtLambda_inv_pow_In  * coeff1 * coeff2 / Lambda0In 
                d_ppbar_dLambdaOut = (dcoeff1 * coeff2 + coeff1 * dcoeff2) * dalpha_dLambdaOut - 0.5 * rtLambda_inv_pow_Out * coeff1 * coeff2 / Lambda0Out
                val -= 0.25 * res_kvec @ np.array([d_ppbar_dLambdaIn,d_ppbar_dLambdaOut]) / res_omega
                result[(newk,newz)] += val
            if Nmin<=tot_pow-2<=Nmax:
                _add_ppbar_bracket_terms((+0.25/res_omega) * coeff1,kz1,coeff2,kz2,result,Lambda0In,Lambda0Out)
    return _consolidate_dictionary_terms(result)

def _add_dicts(*dicts):
    keys = set().union(*[d.keys() for d in dicts])
    dsum = {k:np.sum([d.get(k,0) for d in dicts]) for k in keys}
    return dsum


def _resonance_arguments_of_fixed_order(j,k,N):
    args_dict = df_arguments_dictionary(N)
    args = []
    for N1 in range(k,N+1,2):
        if (N-N1) % 2:
            continue
        nutot = (N-N1)//2
        for arg in args_dict[N1][k]:
            for nuc in _nucombos(nutot):
                js = (j,k - j,*arg)
                args.append((js,nuc))
    return args

def resonant_secular_contribution_dictionary(j,k,Nmin,Nmax,G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out):
    """
    Generate a dictionary containing the secular terms with order between Nmin and 
    Nmax generated by a j:j-k MMR (including its harmonics).

    The secular contributions arise at second order in planet masses.

    Arguments
    ---------
    j : int
      Specifies resonance as the j:j-k MMR.
    k : int
      Order of MMR.
    G : float
      Value of gravitational constant.
    mIn : float
      Inner planet mass parameter
    mOut : float
      Outer planet mass parameter
    MIn : float
      Inner planet stellar mass parameter
    MOut : float
      Outer planet stellar mass parameter
    Lambda0In : float
      Lambda of inner planet for evaluating coefficient.
    Lambda0Out : float
      Lambda of outer planet for evaluating coefficient.
    Nmin : int
      Minimum power (in inclination/eccentricity) of terms to include in result.
    Nmax : int
      Maximum power (in inclination/eccentricity) of terms to include in result.

    Returns
    -------
    result : dict
      A dictionary containing the secular terms with 
      entries in the form
       {(kvec,zvec):Coeff}
    """

    extra_args = (G,mIn,mOut,MIn,MOut,Lambda0In,Lambda0Out)
    # ensure even orders
    Nmin = 2 * (Nmin//2)
    Nmax = 2 * (Nmax//2)
    all_dicts = []
    
    # highest harmonic of j:j-k resonance to include
    nmax = (Nmax + 2) // (2*k)  # Complete secular contribution at desired order
    # loop over harmonics
    for n in range(1,nmax+1):
        j1 = n * j
        k1 = n * k
        res_args = []
    
        # With resonant coefficient p(z,zbar)
        # expanded to order e^(k1 + 2* Mmax + 2), 
        # maximum order secular terms are: 
        #  D_e[e^k1] * D_e [e^(k1 + 2 * Mmax + 2)] = e^Nmax
        # These come from the  ~{pbar,p} / omega_res
        # term in the transformed Hamiltonian so they're
        # subdominant but we'll add them for completeness.
        Mmax = 1 + (Nmax - 2 * k1)//2
        # Mmax = (Nmax - 2 * k1)//2
        for M in range(0,Mmax+1):
            # 
            res_args += _resonance_arguments_of_fixed_order(j1,k1,k1 + 2 * M)
        dres = resonant_terms_list_to_secular_contribution_dictionary(
            res_args,
            j1,
            k1,Nmin,Nmax,
            *extra_args
        )
        all_dicts.append(dres)
    
    return _add_dicts(*all_dicts)



import numpy as np
from sympy import symbols, S, binomial, summation, sqrt, cos, sin, Function,atan2,expand_trig
from celmech.hamiltonian import Hamiltonian
from celmech.disturbing_function import get_fg_coeffs, general_order_coefficient, secular_DF,laplace_B, laplace_coefficient
from celmech.transformations import masses_to_jacobi, masses_from_jacobi
from celmech.resonances import resonance_jk_list
from itertools import combinations
import rebound

def single_true(iterable): # Returns true if only one element in the iterable is set 
    # make generator from iterable setting any zeros as valid entries (otherwise they evaluate to False)
    i = iter([item if item != 0 else True for item in iterable]) # make generator and set zeros to valid inputs
    return any(i) and not any(i) # any(i) True once first valid item found. not any(i) ensures no additional ones exist

class PlanarPoincareParticle(object):
    def __init__(self, m, M, l, gamma, G=1., sLambda=None, sGamma=None, Lambda=None, Gamma=None, a=None, e=None):
        """
        We store the specific Lambda = sqrt(G*M*a) and specific Gamma = sLambda*(1-sqrt(1-e**2)) to support test particles
        """
        if not single_true([sLambda, Lambda, a]):
            raise AttributeError("Can only pass one of Lambda, sLambda (specific Lambda, i.e. per unit mass), or a (semimajor axis)")
        if not single_true([sGamma, Gamma, e]):
            raise AttributeError("Can only pass one of Gamma, sGamma (specific Gamma, i.e. per unit mass), or e (eccentricity)")
        
        if sLambda:
            self.sLambda = sLambda
        elif Lambda:
            try:
                self.sLambda = Lambda/m
            except:
                raise AttributeError("Need to pass specific actions (sLambda and sGamma) or a and e for test particles")
        elif a:
            self.sLambda = np.sqrt(G*M*a)

        if Gamma:
            try:
                sGamma = Gamma/m
            except:
                raise AttributeError("Need to pass specific actions (sLambda and sGamma) or a and e for test particles")
        elif e:
            sGamma = self.sLambda*(1.-np.sqrt(1.-e**2))
        self.sX = np.sqrt(2.*sGamma)*np.cos(gamma) # X per unit sqrt(mass)
        self.sY = np.sqrt(2.*sGamma)*np.sin(gamma)
        self.m = m 
        self.M = M
        self.G = G
        self.l = l
   
    @property
    def X(self):
        return np.sqrt(self.m)*self.sX
    @X.setter
    def X(self, value):
        self.sX = X/np.sqrt(self.m)
    @property
    def Y(self):
        return np.sqrt(self.m)*self.sY
    @Y.setter
    def Y(self, value):
        self.sY = Y/np.sqrt(self.m)
    @property
    def Lambda(self):
        return self.m*self.sLambda
    @Lambda.setter
    def Lambda(self, value):
        self.sLambda = value/self.m
    @property
    def Gamma(self):
        return self.m*(self.sX**2+self.sY**2)/2.
    @Gamma.setter
    def Gamma(self, value):
        self.sGamma = value/self.m
    @property
    def sGamma(self):
        return (self.sX**2+self.sY**2)/2.
    @property
    def gamma(self):
        return np.arctan2(self.sY, self.sX)
    @property
    def a(self):
        return self.sLambda**2/self.G/self.M
    @property
    def e(self):
        GbyL = self.sGamma/self.sLambda
        if 1-(1.-GbyL)*(1.-GbyL) < 0:
            raise AttributeError("sGamma:{0}, sLambda:{1}, GbyL:{2}, val:{3}".format(self.sGamma, self.sLambda, GbyL, 1-(1.-GbyL)*(1.-GbyL)))
        return np.sqrt(1 - (1-GbyL)*(1-GbyL))
    @property
    def pomega(self):
        return -self.gamma
    @property
    def n(self):
        return np.sqrt(self.G*self.M/self.a**3)

class PlanarPoincare(object):
    def __init__(self, G, poincareparticles=[]):
        self.G = G
        self.particles = [PlanarPoincareParticle(m=np.nan, M=np.nan, G=np.nan, l=np.nan, gamma=np.nan, sLambda=np.nan, sGamma=np.nan)] # dummy particle for primary
        try:
            for p in poincareparticles:
                self.add(m=p.m, sLambda=p.sLambda, l=p.l, sGamma=p.sGamma, gamma=p.gamma, M=p.M)
        except TypeError:
            raise TypeError("poincareparticles must be a list of PlanarPoincareParticle objects")

    @classmethod
    def from_Simulation(cls, sim, average=True):
        masses = [p.m for p in sim.particles]
        mjac, Mjac = masses_to_jacobi(masses)
        pvars = PlanarPoincare(sim.G)
        ps = sim.particles
        o = sim.calculate_orbits(jacobi_masses=True)
        for i in range(1,sim.N-sim.N_var):
            M = Mjac[i]
            m = mjac[i]
            orb = o[i-1]
            if orb.a <= 0. or orb.e >= 1.:
                raise AttributeError("Celmech error: PlanarPoincare.from_Simulation only support elliptical orbits. Particle {0}'s (jacobi) a={1}, e={2}".format(i, orb.a, orb.e))
            sLambda = np.sqrt(sim.G*M*orb.a)
            sGamma = sLambda*(1.-np.sqrt(1.-orb.e**2))
            pvars.add(m=m, sLambda=sLambda, l=orb.l, sGamma=sGamma, gamma=-orb.pomega, M=M)
        if average is True:
            pvars.average_synodic_terms()
        return pvars

    def to_Simulation(self, masses=None, average=True):
        ''' 
        if masses is None, will calculate physical masses from the jacobi ones.
        if masses is a list, will use those as the physical masses.
        '''

        if average is True:
            self.average_synodic_terms(inverse=True)

        if not masses:
            mjac = [p.m for p in self.particles]
            Mjac = [p.M for p in self.particles]
            masses = masses_from_jacobi(mjac, Mjac)

        sim = rebound.Simulation()
        sim.G = self.G
        sim.add(m=masses[0])
        ps = self.particles
        #print(mjac, Mjac, masses)
        for i in range(1, self.N):
            sim.add(m=masses[i], a=ps[i].a, e=ps[i].e, pomega=-ps[i].gamma, l=ps[i].l, jacobi_masses=True)
        sim.move_to_com()
        return sim
    
    def add(self, **kwargs):
        self.particles.append(PlanarPoincareParticle(G=self.G, **kwargs))

    def copy(self):
        return PlanarPoincare(self.G, self.particles[1:self.N])

    def average_resonant_terms(self, i1=1, i2=2, deltaP=0.03, exclude=[], order=2, inverse=False):
        """
        Do a canonical transformation to correct the Lambdas for the fact that we have implicitly
        averaged over all the resonant terms we do not include in the Hamiltonian.
        """
        ps = self.particles
        m1 = ps[i1].m
        m2 = ps[i2].m
        n1 = ps[i1].n
        n2 = ps[i2].n
        Pratio = n2/n1 # P1/P2
        alpha = ps[i1].a/ps[i2].a
        e1 = ps[1].e
        e2 = ps[2].e
        l1 = ps[i1].l
        l2 = ps[i2].l
        gamma1 = ps[i1].gamma
        gamma2 = ps[i2].gamma
        G = self.G
        prefac = G/ps[i2].a

        sum1 = 0.
        sum2 = 0.
        prevsum=0.
        jklist = resonance_jk_list(Pratio-deltaP, min(Pratio+deltaP, 0.995), order)
        for j,k in jklist:
            if [j,k] in exclude:
                continue
            prefac1 = (j-k)/(j*n2 - (j-k)*n1)
            prefac2 = j/(j*n2 - (j-k)*n1)
            theta = j*l2 - (j-k)*l1
            for l in range(k+1): # 0 to k inclusive
                Cjkl = general_order_coefficient(j,k,l,alpha)
                Bjkl = Cjkl*e1**l*e2**(k-l) 
                cosine = np.cos(theta + l*gamma1 - (l-k)*gamma2)
                sum1 += prefac1*Bjkl*cosine
                sum2 += prefac2*Bjkl*cosine
                denom = 1.-float(j-k)/j*n1/n2
                #print(j,k,(sum1-prevsum)*prefac*m2/ps[i1].sLambda, denom, 1./j*m2/ps[i2].M/denom)
            prevsum=sum1

        sum1 *= prefac
        sum2 *= prefac
        print(sum1*m2/ps[i1].sLambda, sum2*m1/ps[i2].sLambda)
        if inverse is True:
            sum1 *= -1
            sum2 *= -1
        ps[i1].sLambda += m2*sum1
        ps[i2].sLambda -= m1*sum2

    def average_synodic_terms(self, inverse=False):
        """
        Do a canonical transformation to correct the Lambdas for the fact that we have implicitly
        averaged over all the synodic terms we do not include in the Hamiltonian.
        """
        corrpvars = self.copy() # so we use original values when planet appears in more than one pair
        pairs = combinations(range(1,self.N), 2)
        #TODO assumes particles ordered going outward so a1 < a2 always. Sort first?
        for i1, i2 in pairs:
            ps = self.particles
            m1 = ps[i1].m
            m2 = ps[i2].m
            deltalambda = ps[i1].l-ps[i2].l
            G = self.G

            prefac = G/ps[i2].a/(ps[i1].n-ps[i2].n) 
            alpha = ps[i1].a/ps[i2].a
            summation = (1. + alpha**2 - 2*alpha*np.cos(deltalambda))**(-0.5)
            s = prefac*(alpha*np.cos(deltalambda)-summation+laplace_coefficient(0.5, 0, 0, alpha)/2.)
            if inverse is True:
                s *= -1
            corrpvars.particles[i1].sLambda += m2*s # prefac*m1*m2*s/m1 (sLambda=Lambda/m)
            corrpvars.particles[i2].sLambda -= m1*s
        
        for i, p in enumerate(self.particles):
            p.sLambda = corrpvars.particles[i].sLambda

    @property
    def N(self):
        return len(self.particles)

class PlanarPoincareHamiltonian(Hamiltonian):
    def __init__(self, pvars):
        Hparams = {symbols('G'):pvars.G}
        pqpairs = []
        ps = pvars.particles
        H = S(0) 
        for i in range(1, pvars.N):
            pqpairs.append(symbols("X{0}, Y{0}".format(i))) 
            pqpairs.append(symbols("Lambda{0}, lambda{0}".format(i))) 
            Hparams[symbols("m{0}".format(i))] = ps[i].m
            Hparams[symbols("M{0}".format(i))] = ps[i].M
            H = self.add_Hkep_term(H, i)
        self.resonance_indices = []
        super(PlanarPoincareHamiltonian, self).__init__(H, pqpairs, Hparams, pvars) 
        # Create polar hamiltonian for display/familiarity for user
        self.Hpolar = S(0)
        for i in range(1, pvars.N):
            self.Hpolar = self.add_Hkep_term(self.Hpolar, i)

        # Don't re-compute secular DF terms, save a dictionary
        # of secular DF expansions indexed by order:
        self.secular_terms = {0:S(0)}
    
    @property
    def particles(self):
        return self.state.particles

    @property
    def N(self):
        return len(self.particles)
    
    def state_to_list(self, state):
        ps = state.particles
        vpp = 4 # vars per particle
        y = np.zeros(vpp*(state.N-1)) # remove padded 0th element in ps for y
        for i in range(1, state.N):
            y[vpp*(i-1)] = ps[i].X
            y[vpp*(i-1)+1] = ps[i].Y
            y[vpp*(i-1)+2] = ps[i].Lambda
            y[vpp*(i-1)+3] = ps[i].l 
        return y
    def set_secular_mode(self):
        # 
        state = self.state
        for i in range(1,state.N):
            Lambda0,Lambda = symbols("Lambda{0}0 Lambda{0}".format(i))
            self.H = self.H.subs(Lambda,Lambda0)
            self.Hparams[Lambda0] = state.particles[i].Lambda
        self._update()

    def update_state_from_list(self, state, y):
        ps = state.particles
        vpp = 4 # vars per particle
        for i in range(1, state.N):
            ps[i].sX = y[vpp*(i-1)]/np.sqrt(ps[i].m)
            ps[i].sY = y[vpp*(i-1)+1]/np.sqrt(ps[i].m)
            ps[i].sLambda = y[vpp*(i-1)+2]/ps[i].m
            ps[i].l = y[vpp*(i-1)+3]
    
    def add_Hkep_term(self, H, index):
        """
        Add the Keplerian component of the Hamiltonian for planet ''.
        """
        G, M, m, Lambda = symbols('G, M{0}, m{0}, Lambda{0}'.format(index))
        #m, M, mu, Lambda, lam, Gamma, gamma = self._get_symbols(index)
        H +=  -G**2*M**2*m**3 / (2 * Lambda**2)
        return H
    
    def add_secular_terms(self, order=2,fixed_Lambdas=True, indexIn=1, indexOut=2):
        G = symbols('G')
        mOut,MOut,LambdaOut,lambdaOut,GammaOut,gammaOut,XOut,YOut = symbols('m{0},M{0},Lambda{0},lambda{0},Gamma{0},gamma{0},X{0},Y{0}'.format(indexOut)) 
        mIn,MIn,LambdaIn,lambdaIn,GammaIn,gammaIn,XIn,YIn = symbols('m{0},M{0},Lambda{0},lambda{0},Gamma{0},gamma{0},X{0},Y{0}'.format(indexIn)) 

        
        eIn,eOut,gammaIn,gammaOut = symbols("e e' gamma gamma'") 
        hIn,kIn,hOut,kOut=symbols("h,k,h',k'")
        # Work smarter not harder! Use an expression it is already available...
        if order not in self.secular_terms.keys():
            print("Computing secular expansion to order %d..."%order)
            subdict={eIn:sqrt(hIn*hIn + kIn*kIn), eOut:sqrt(hOut*hOut + kOut*kOut),gammaIn:atan2(-1*kIn,hIn),gammaOut:atan2(-1*kOut,hOut),}
            self.secular_terms[order] = secular_DF(eIn,eOut,gammaIn,gammaOut,order).subs(subdict,simultaneous=True)

        exprn = self.secular_terms[order]
        salpha = S("alpha{0}{1}".format(indexIn,indexOut))
        if order==2:
            subdict = { hIn:XIn/sqrt(LambdaIn) , kIn:(-1)*YIn/sqrt(LambdaIn), hOut:XOut/sqrt(LambdaOut) , kOut:(-1)*YOut/sqrt(LambdaOut), S("alpha"):salpha}
        else:
            # fix this to include higher order terms
            subdict = { hIn:XIn/sqrt(LambdaIn) , kIn:(-1)*YIn/sqrt(LambdaIn), hOut:XOut/sqrt(LambdaOut) , kOut:(-1)*YOut/sqrt(LambdaOut), S("alpha"):salpha}

        alpha = self.particles[indexIn].a/self.particles[indexOut].a
        self.Hparams[salpha] = alpha
        self.Hparams[Function('b')] = laplace_B
        exprn = exprn.subs(subdict)
        # substitute a fixed value for Lambdas in DF terms

        prefactor = -G**2*MOut**2*mOut**3 *( mIn / MIn) / (LambdaOut**2)
        exprn = prefactor * exprn
        if fixed_Lambdas:
            LambdaIn0,LambdaOut0=symbols("Lambda{0}0 Lambda{1}0".format(indexIn,indexOut))
            self.Hparams[LambdaIn0]=self.state.particles[indexIn].Lambda
            self.Hparams[LambdaOut0]=self.state.particles[indexOut].Lambda
            exprn = exprn.subs([(LambdaIn,LambdaIn0),(LambdaOut,LambdaOut0)])
        self.H += exprn
        self._update()

    def add_all_resonance_subterms(self, j, k, indexIn=1, indexOut=2):
        """
        Add all the terms associated the j:j-k MMR between planets 'indexIn' and 'indexOut'.
        Inputs:
        indexIn     -    index of the inner planet
        indexOut    -    index of the outer planet
        j           -    together with k specifies the MMR j:j-k
        k           -    order of the resonance
        """
        for l in range(k+1):
            self.add_single_resonance(j=j, k=k, l=l, indexIn=indexIn, indexOut=indexOut) 

    def add_single_resonance(self, j, k, l, indexIn=1, indexOut=2):
        """
        Add a single term associated the j:j-k MMR between planets 'indexIn' and 'indexOut'.
        Inputs:
        indexIn     -   index of the inner planet
        indexOut    -   index of the outer planet
        j           -   together with k specifies the MMR j:j-k
        k           -   order of the resonance
        l           -   picks out the eIn^(l) * eOut^(k-l) subterm
        """
        # Canonical variables
        assert indexOut == indexIn + 1,"Only resonances for adjacent pairs are currently supported"
        G = symbols('G')
        mIn,MIn,LambdaIn,lambdaIn,XIn,YIn = symbols('m{0},M{0},Lambda{0},lambda{0},X{0},Y{0}'.format(indexIn)) 
        mOut,MOut,LambdaOut,lambdaOut,XOut,YOut = symbols('m{0},M{0},Lambda{0},lambda{0},X{0},Y{0}'.format(indexOut)) 
        #mIn, MIn, muIn, LambdaIn,lambdaIn,XIn,YIn = self._get_symbols(indexIn)
        #mOut, MOut, muOut, LambdaOut,lambdaOut,XOut,YOut = self._get_symbols(indexOut)
        
        # Resonance index
        assert l<=k, "Invalid resonance term, l must be less than or equal to k."
        alpha = self.particles[indexIn].a/self.state.particles[indexOut].a
	
        # Resonance components
        #
        Cjkl = symbols( "C_{0}\,{1}\,{2}".format(j,k,l) )
        self.Hparams[Cjkl] = general_order_coefficient(j,k,l,alpha)
        #
        
        i=symbols('i')
        lBy2 = int(np.floor(l/2.))
        k_l = k-l
        k_lBy2 = int(np.floor((k_l)/2.))

        if l> 0:
            z_to_p_In_re = summation( binomial(l,(2*i)) * XIn**(l-(2*i)) * YIn**(2*i) * (-1)**(i) , (i, 0, lBy2 ))
            z_to_p_In_im = summation( binomial(l,(2*i+1)) * XIn**(l-(2*i+1)) * YIn**(2*i+1) * (-1)**(i) , (i, 0, lBy2 ))
        else:
            z_to_p_In_re = 1
            z_to_p_In_im = 0
        if k_l>0:
            z_to_p_Out_re = summation( binomial(k_l,(2*i)) * XOut**(k_l-(2*i)) * YOut**(2*i) * (-1)**(i) , (i, 0, k_lBy2 ))
            z_to_p_Out_im = summation( binomial(k_l,(2*i+1)) * XOut**(k_l-(2*i+1)) * YOut**(2*i+1) * (-1)**(i) , (i, 0, k_lBy2 ))
        else:
            z_to_p_Out_re = 1
            z_to_p_Out_im = 0
        reFactor = (z_to_p_In_re * z_to_p_Out_re - z_to_p_In_im * z_to_p_Out_im) / sqrt(LambdaIn)**l / sqrt(LambdaOut)**k_l
        imFactor = (z_to_p_In_im * z_to_p_Out_re + z_to_p_In_re * z_to_p_Out_im) / sqrt(LambdaIn)**l / sqrt(LambdaOut)**k_l
        # Keep track of resonances
        self.resonance_indices.append((indexIn,indexOut,(j,k,l)))
        # Update internal Hamiltonian
        prefactor = -G**2*MOut**2*mOut**3 *( mIn / MIn) / (LambdaOut**2)
        costerm = cos(j * lambdaOut  - (j-k) * lambdaIn )
        sinterm = sin(j * lambdaOut  - (j-k) * lambdaIn )
        
        self.H += prefactor * Cjkl * (reFactor * costerm - imFactor * sinterm)
        self._update()
        
        # update polar Hamiltonian
        GammaIn,gammaIn,GammaOut,gammaOut = symbols('Gamma{0},gamma{0},Gamma{1},gamma{1}'.format(indexIn, indexOut))
        eccIn = sqrt(2*GammaIn/LambdaIn)
        eccOut = sqrt(2*GammaOut/LambdaOut)
        #
        costerm = cos( j * lambdaOut - (j-k) * lambdaIn + l * gammaIn + (k-l) * gammaOut )
        #
        self.Hpolar += prefactor * Cjkl * (eccIn**l) * (eccOut**(k-l)) * costerm



import warnings
from scipy.linalg import solve as lin_solve
from numpy import sqrt
import numpy as np
_rt2 = sqrt(2)
_rt2_inv = 1 / _rt2 
_ImplicitMidpoint = {
        'a':np.array([[0.5]]),
        'c':np.array([0.5]),
        'b':np.array([1])
        }
_LobattoIIIB = {
    'a':np.array([
        [1/6, -1/6, 0],
        [1/6, 1/3, 0],
        [1/6, 5/6, 0]
    ]),
    'c':np.array([0, 1/2, 1]),
    'b':np.array([1/6, 2/3, 1/6])
}

_GL4 = {
    'a':np.array([
        [1/4, 1/4-1/6*sqrt(3)],
        [ 1/4+1/6*sqrt(3), 1/4]
    ]),
    'c':np.array([1/2 - 1/6*sqrt(3),1/2 + 1/6*sqrt(3)]),
    'b':np.array([1/2,1/2])
}

_GL6 = {
    'a':np.array([
        [5/36,2/9-1/15*sqrt(15),5/36-1/30*sqrt(15)],
        [5/36 + 1/24*sqrt(15),2/9,5/36 - 1/24*sqrt(15)],
        [5/36 + 1/30*sqrt(15),   2/9 + 1/15*sqrt(15),5/36]
    ]),
    'c':np.array([1/2 - 1/10*sqrt(15), 1/2, 1/2 + 1/10*sqrt(15)]),
    'b':np.array([5/18,  4/9,  5/18])
}

# Explicit methods
_RK4 = {
        'a':np.array([
            [0.,0.,0.,0.],
            [0.5,0.,0.,0.],
            [0,0.5,0.,0.],
            [0,0,1,0]
            ]),
        'c':np.array([0.,0.5,0.5,1.]),
        'b':np.array([1/6,1/3,1/3,1/6])
}

_ExplicitMidpoint = {
        'a':np.array([
            [0,0],
            [0.5,0]
            ]),
        'b':np.array([0,1]),
        'c':np.array([0,0.5])
        }

# all methods
_rk_methods = {
        'ImplicitMidpoint':_ImplicitMidpoint,
        'LobattoIIIB':_LobattoIIIB,
        'GL4':_GL4,
        'GL6':_GL6,
        'RK4':_RK4,
        'ExplicitMidpoint':_ExplicitMidpoint
}

class RKIntegrator():
    def __init__(self,f,f_and_Df,Ndim,dt,rtol,atol,rk_method,rk_root_method,max_iter):
        
        self.f = f
        self.f_and_Df = f_and_Df
        self.dt = dt
        self.Ndim = Ndim
        self.rtol = rtol
        self.atol = atol
        self.rk_root_method = rk_root_method
        self.rk_method = rk_method
        self.max_iter = max_iter
        
            
        # Set tolerance and iteration parameters
        tols_allowed = atol >=0 and rtol >=0
        tols_allowed = tols_allowed and (atol > 0 or rtol > 0)
        assert tols_allowed, "Tolerances must be non-negative and at least one tolerance must be positive." 
        
    @property
    def dt(self):
        return self._dt
    
    @dt.setter
    def dt(self,dt):
        self._dt = dt
        
    @property
    def rk_method(self):
        return self._rkmethod

    @rk_method.setter
    def rk_method(self,rkmethod):
        if rkmethod in _rk_methods.keys():
            self._rkmethod = rkmethod
            self._rk_tableau=_rk_methods[rkmethod]
            self.rk_a= self._rk_tableau['a']
            self.rk_b= self._rk_tableau['b']
            self.rk_c= self._rk_tableau['c']
            self.rk_s = len(self.rk_c)
        else:
            methods_list = "\n".join(["\t{:s}".format(method) for method in _rk_methods.keys()])
            raise ValueError("'rkmethod' must be one of:\n" + methods_list) 

        if self.rk_root_method=='explicit':
            for i,ai  in enumerate(self.rk_a):
                assert np.alltrue(ai[i:]==0), "RK tableau cannot be solved explicitly"

    @property 
    def rk_root_method(self):
        return self._rk_root_method
    
    @rk_root_method.setter
    def rk_root_method(self,rk_root_method):
        if rk_root_method == 'Newton':
            self.rk_step = self._implicit_rk_step_newton
        elif rk_root_method == 'quasi-Newton':
            self.rk_step = self._implicit_rk_step_quasi_newton
        elif rk_root_method == 'fixed_point':
            self.rk_step = self._implicit_rk_step_fixed_point
        elif rk_root_method == 'explicit':
            self.rk_step = self._explicit_rk_step
        else:
            raise ValueError("'rk_root_method' must be either 'Newton', 'quasi-Newton', 'fixed_point', or 'explicit'")
        self._rk_root_method = rk_root_method

    def _implicit_step_root_eqn(self,K,y):
        h = self.dt
        a = self.rk_a
        s = self.rk_s
        Ndim = self.Ndim
        f_and_Df = self.f_and_Df
        k = K.reshape((s,Ndim))
        knew = np.zeros((s,Ndim))
        Dkdy = np.zeros((s,Ndim,Ndim))
        ytemps = y + h * a @ k
        Imtrx = np.eye(s * Ndim)
        for i,ytemp in enumerate(ytemps):
            knew[i],Dkdy[i] = f_and_Df(ytemp)
        fOfK=np.hstack(knew)
        g = K - fOfK
        Dg = Imtrx - h * np.block([[ a[i,j] * Dkdy[i] for j in range(s)] for i in range(s)])
        return g,Dg

    def _implicit_rk_step_newton(self,y):
        """
        Advance ODE for input y for a timestep h
        using an implicit Runge-Kutta method defined by the
        Butcher tableau [a,b,c].

        Arguments
        ---------
        qp_vec
        """
        h = self.dt
        a = self.rk_a
        b = self.rk_b
        c = self.rk_c
        s = self.rk_s
        Ndim = self.Ndim
        f = self.f
        f_and_Df = self.f_and_Df
        
        max_iter = self.max_iter
        k = np.zeros((s,Ndim))
        Dkdy = np.zeros((s,Ndim,Ndim))
        rtol = self.rtol
        atol = self.atol
        Imtrx = np.eye(s * Ndim)
        #
        
        # Generate initial guesses from second-order approx.
        ####################################################
        # This appears to be sligtly faster than
        # than generating guesses via RK4, as in
        #  _implicit_rk_step_fix_point, but I haven't
        # tested it extensively.
        ####################################################
        ktemp,Dktemp = f_and_Df(y)
        d2ydt2 = Dktemp @ ktemp
        k = np.array([ktemp + ci*h*d2ydt2 for ci in c])
        # Main loop
        K = np.hstack(k)
        for itr in range(max_iter):
            g,Dg = self._implicit_step_root_eqn(K,y)
            dK = -1 * lin_solve(Dg,g)
            K += dK
            if np.alltrue( np.abs(dK) < rtol * np.abs(K) + atol ):
                break
        else:
            warnings.warn("'implicit_rk_step' reached maximum number of iterations ({})".format(max_iter))
        k = K.reshape(s,Ndim) 
        ynew = y + h * b @ k
        return ynew
    
    def _rk4_step(self,y,ydot,h):
        f = self.f
        h_by_2 = 0.5 * h
        k1 = ydot
        y2 = y + h_by_2 * k1
        k2 = f(y2)
        y3 = y + h_by_2 * k2
        k3 = f(y3)
        y4 = y + h * k3
        k4 = f(y4)
        yout = y + h * (k1 + 2*k2 + 2*k3 + k4) / 6.
        return yout,f(yout)

    def _explicit_rk_step(self,qp_vec):
        h = self._dt
        a = self.rk_a
        b = self.rk_b
        s = self.rk_s
        f = self.f
        y = qp_vec
        k = np.zeros((s,self.Ndim))
        for i,ai in enumerate(a):
            k[i] = f(y + h * ai @ k)
        ynew = y + h * b @ k
        return ynew

    def _implicit_rk_step_fixed_point(self,qp_vec):
        """
        Advance ODE for input qpve for a timestep h
        using an implicit Runge-Kutta method defined by the
        Butcher tableau [a,b,c].

        Arguments
        ---------
        qp_vec
        """
        h = self._dt
        a = self.rk_a
        b = self.rk_b
        c = self.rk_c
        f = self.f
        y = qp_vec
        ktemp = f(qp_vec)
        max_iter = self.max_iter
        k = np.zeros((self.rk_s,self.Ndim))
        rtol = self.rtol
        atol = self.atol
        # 
        for i,ci in enumerate(c):
            if ci == 0:
                k[i,:] = ktemp 
            else:
                _,k[i,:] = self._rk4_step(y,ktemp,ci*h)
        for itr in xrange(max_iter):
            k_old = k.copy()
            ytemps = y + h * a @ k 
            k = np.array([f(ytemp) for ytemp in ytemps])
            delta_ks = np.abs(k-k_old)
            if np.alltrue( delta_ks < rtol * np.abs(k_old) + atol ):
                break
        else:
            warnings.warn("'implicit_rk_step' reached maximum number of iterations ({})".format(max_iter))
        ynew = y + h * b @ k
        return ynew
    
    def _implicit_rk_step_quasi_newton(self,qp_vec):
        """
        Advance ODE for input qpve for a timestep h
        using an implicit Runge-Kutta method defined by the
        Butcher tableau [a,b,c].

        Arguments
        ---------
        qp_vec
        """
        h = self._dt
        a = self.rk_a
        b = self.rk_b
        c = self.rk_c
        s = self.rk_s
        Ndim = self.Ndim
        f = self.f
        f_and_Df = self.f_and_Df
        y = qp_vec
        Imtrx = np.eye(s * Ndim)
        max_iter = self.max_iter
        rtol = self.rtol
        atol = self.atol

        # set up method
        k = np.zeros((s,Ndim))
        ktemp,Dkdy0 = f_and_Df(y)
        d2ydt2 = Dkdy0 @ ktemp
        k = np.array([ktemp + ci*h*d2ydt2 for ci in c])
        Dgdy0 = Imtrx - h * np.block([[ a[i,j] * Dkdy0 for j in range(s)] for i in range(s)])
        Dgdy0_inv = np.linalg.inv(Dgdy0)
        # Main loop
        for itr in range(max_iter):
            ytemps = y + h * a @ k
            g = np.hstack(k-np.array([f(ytemp) for ytemp in ytemps]))
            dK = -1 * Dgdy0_inv @ g
            dk = dK.reshape(s,Ndim)
            k += dk
            if np.alltrue( np.abs(dk) < rtol * np.abs(k) + atol ):
                break
        else:
            warnings.warn("'implicit_rk_step' reached maximum number of iterations ({})".format(max_iter))
        ynew = y + h * b @ k
        return ynew

